/*-------------------------------------------------------------------------*/ 
/*                                                                         */
/*                   G.723.1 Channel Decoder Simulation V3.1               */
/*                                                                         */
/*-------------------------------------------------------------------------*/
/*                                                                         */
/*-------------------------------------------------------------------------*/
/* (C) Copyright 1995-1996, Bosch Telecom. All rights reserved.            */
/* Contact: Joerg-Martin Mueller, Bosch Telecom, Email: jmm@bk.bosch.de    */
/*-------------------------------------------------------------------------*/


/*___________________________________________________________________________
 |                                                                           |
 | 	Include-Files                                                        |
 |___________________________________________________________________________|
*/
#include "defines.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

/* G.723.1 Includes */
#include "typedef.h"
#include "basop.h"
#include "cst_lbc.h"
#include "tab_lbc.h"
#include "lbccodec.h"
#include "coder.h"
#include "decod.h"
#include "exc_lbc.h"
#include "util_lbc.h"

#define   setbit( x, n )    (( x ) |=   ( 1 << ( n )))
#define   clrbit( x, n )    (( x ) &= (~( 1 << ( n ))))


/*___________________________________________________________________________
 |                                                                           |
 |  	local Functions                                                      |
 |___________________________________________________________________________|
*/
static void  GetChannelBitrates(char* ConfigName);
static void CrcDecoder(Word16 info_bit_r[], Word16 *bad_frame, Word16 G723_mode_r);
static void bit_deallocation(Word16 info_bit_r[],  char InfoBitStream_r[], 
                             Word16 G723_mode_r );
static void  EndTrellisReorderDec (Word16 info_bit[], Word16 info_bit_soft[], Word16 bit_class[], 
                                   Word16 punctur[5][3], Word16 mode);
static void ChannelDecoder(Word16 chan_bit_r[], char InfoBitStream_r[]);
static void viterbi( Word16 nbit_class[], Word16 punctur[5][3], 
    Word16 chan_bit_r[], Word16 info_bit_r[], Word16 info_bit_r_soft[]);
static void hamming_distance( Word16 chan_bit_np[], Word16 metric[], Word16 temp[] );
static void metric_update( Word16 chan_bit_np[], Word16 metric[], Word16 *surv_mem );
static void hamming_distance( Word16 chan_bit_np[], Word16 metric[], Word16 temp[] );
static void metric_update( Word16 chan_bit_np[], Word16 metric[], Word16 *surv_mem );
static void ChannelCodecInit(Word16 SpeechCodecMode,Word16 ChannelBitrate,
                             Word16 bit_class[],Word16 punctur[5][3]);
static Word16 GetRate (Word16 rbits, Word16 weight, Word16 bit_class, Word16 inv_bit_class, Word16 rnd);
static bitrate_test( Word16 nbit_class[], Word16 punctur[5][3], Word16 last );
static void bitrate_adapt_high (Word16 bit_class[],Word16 punctur[5][3],
                                Word16 cbits, Word16 r[], Word16 last);
static void bitrate_adapt_low (Word16 bit_class[],Word16 punctur[5][3],
                                Word16 cbits, Word16 r[], Word16 last);
static void SoftBFI (Word16 bit_class[], Word16 punctur[5][3], 
           Word16 info_bit_r_soft[], Word16 bad_frame_soft[], Word16 bad_frame,
Word16 G723_mode_r);
static void ConfigDecode (Word16 chan_bit_r[], Word16 *mode_r, Word16 *err_conf);
static void speech_decoder_adaptation(char CInfoBitStream[],char InfoBitStream[],
                                       Word16 IndexError[]);
static LINEDEF  Line_Unpk_Channel( char *Vinp, Word16 IndexError[] );
static void LSP_Reorder_decoder (LINEDEF *Line);
static void ModeChangeAccept (Word16 mode_r, Word16 err_conf, Word16 *ModeAccept);
static void DecodeChannelBitrate(Word16 CC_mode_r,Word16 G723_mode_r, Word16 *SysChannelBitrate_r);
static int Strincmp( const char *s, const char *t, size_t max );
static FILE *OpenBinfile( char *name, char *mode );
static void  Init_DecoderState (void);

/*___________________________________________________________________________
 |                                                                           |
 |  	Constants & Globals                                                  |
 |___________________________________________________________________________|
*/
static char   outtext[]                = " Decoding Frame: %4d\r";
enum  Wmode    WrkMode = Both ;
enum  Crate    WrkRate;
Flag     UseHp = True;
Flag     UsePf;
Flag     UseVx;
Word32   count_clip;
Word16   MAX_CHAN_BIT = 684;



/* Definition of CRC Check Code for error detection (see section 2.4) */
/* ------------------------------------------------------------------ */
#define   CrcLen  5
#define   CrcGen  18
static const Word16   CrcWindow53   = 34; /* number of bits of class 1 for crc */
static const Word16   CrcWindow63   = 44; /* number of bits of class 1 for crc */
static const Word16   CrcWindowSID  = 30; /* number of bits of class 1 for crc */

/* Definition of low rate configuration */
/* ------------------------------------ */
#define   NBits5_3       168  /* Info bits with configuration and UB's */
#define   NInfoBits5_3   162  /* Info bits without configuration and UB's */

/* Definition of high rate configuration */
/* ------------------------------------- */
#define   NBits6_3       200  /* Info bits with configuration and UB's */
#define   NInfoBits6_3   193  /* Info bits without configuration and UB's */

/* Definition of SID configuration */
/* ------------------------------- */
#define   NBitsSID        32  /* Info bits with configuration and UB's */
#define   NInfoBitsSID    30  /* Info bits without configuration and UB's */

/* Definition of BCH Code for configuration bits (see section 2.5.4) */
/* ----------------------------------------------------------------- */
#define   BCHLen  13
#define   NumBCHWords  8

/* Channel bitrates (are read from file) */
/* ------------------------------------- */
static Word16   ChannelBitrate_53[2];
static Word16   ChannelBitrate_63[2];
static Word16   ChannelBitrate_SID[2];

/*___________________________________________________________________________
 |                                                                           |
 |  	Subroutines                                                          |
 |___________________________________________________________________________|
*/


static void DecodeChannelBitrate(Word16 CC_mode_r,Word16 G723_mode_r, 
                                 Word16 *SysChannelBitrate_r)
{
   Word16   CC_Bitrate[2],i;

   if (G723_mode_r == 0){
      for (i=0;i<2;i++) CC_Bitrate[i] = ChannelBitrate_63[i];
   }
   if (G723_mode_r == 1){
      for (i=0;i<2;i++) CC_Bitrate[i] = ChannelBitrate_53[i];
   }
   if (G723_mode_r == 2){
      for (i=0;i<2;i++) CC_Bitrate[i] = ChannelBitrate_SID[i];
   }

   *SysChannelBitrate_r = CC_Bitrate[CC_mode_r];

   return;
}


static void  EndTrellisReorderDec (Word16 info_bit[], Word16 info_bit_soft[], 
                         Word16 bit_class[], Word16 punctur[5][3], Word16 mode)
{
   Word16 last,i,ptr_end,tmp,limit;

    if( (mode & 0x3) == 0) limit=CrcWindow63+CrcLen+20;
    if( (mode & 0x3) == 1) limit=CrcWindow53+CrcLen+20;
    if( (mode & 0x3) == 2) limit=CrcWindowSID+CrcLen+20;


    /* Check which class is the last protected one */
   last=0;
   for (i=1;i<5;i++){
      if (punctur[i][0] != 0) last=i;
   }

   /* Determine number of protected bits */
   ptr_end = 0;
   for (i=0;i<=last;i++) ptr_end = ptr_end + bit_class[i];
   ptr_end = ptr_end - 4;  /* remove tail bits */

   if (ptr_end >= limit) {
      ptr_end--;
      for (i=0;i<10;i++){
         tmp                    = info_bit[ptr_end-i];
         info_bit[ptr_end-i]    = info_bit[ptr_end-19+i];
         info_bit[ptr_end-19+i] = tmp;

         tmp                         = info_bit_soft[ptr_end-i];
         info_bit_soft[ptr_end-i]    = info_bit_soft[ptr_end-19+i];
         info_bit_soft[ptr_end-19+i] = tmp;
      }
   }
   return;
}



/*_________________________________________________________________________________
 |                                                                                 |
 | Determine the parameters of the channel encoder given the channel bitrate       |
 |               (section 2.5.1)                                                   |
 |_________________________________________________________________________________|
*/
static void ChannelCodecInit(Word16 SpeechCodecMode,Word16 ChannelBitrate,
                             Word16 bit_class[],Word16 punctur[5][3])

{
/* Definition of puncturing matrix for convolutional code (see Table C4) */
/* --------------------------------------------------------------------- */
   static const Word16 PuncturTable[25][3] = {{0x000,0x000,0x000},
                                            {0xd6f,0x690,0x000},
                                            {0xd7f,0x690,0x000},
                                            {0xd7f,0x691,0x000},
                                            {0xd7f,0x695,0x000},
                                            {0xdff,0x695,0x000},
                                            {0xfff,0x695,0x000},
                                            {0xfff,0x69d,0x000},
                                            {0xfff,0x6dd,0x000},
                                            {0xfff,0x6df,0x000},
                                            {0xfff,0x7df,0x000},
                                            {0xfff,0x7ff,0x000},
                                            {0xfff,0xfff,0x000},
                                            {0xfff,0xfff,0x001},
                                            {0xfff,0xfff,0x009},
                                            {0xfff,0xfff,0x109},
                                            {0xfff,0xfff,0x309},
                                            {0xfff,0xfff,0x329},
                                            {0xfff,0xfff,0x729},
                                            {0xfff,0xfff,0x72d},
                                            {0xfff,0xfff,0x72f},
                                            {0xfff,0xfff,0x7af},
                                            {0xfff,0xfff,0x7bf},
                                            {0xfff,0xfff,0x7ff},
                                            {0xfff,0xfff,0xfff}};

/* Definition of low rate configuration (see Table C5a)         */
/* Note that the weight is in 16 bit representation and negated */
/* ------------------------------------------------------------ */
   static const Word16   BitClass53[5]    = { 34+CrcLen, 40, 40, 40+4, 8 };
   static const Word16   Weight53[5]      = { -7864, -10158, -7864, -6882, 0}; 

   static const Word16   InvBitClass53[4] = {26887, 26214, 26214, 26214}; /* 2^20/BitClass53[i]) */
   static const Word16   inv_b0p4_53      = 24386;                        /* 2^20/(BitClass53[0]+4)*/

/* Definition of high rate configuration (see Table C5a)        */
/* Note that the weight is in 16 bit representation and negated */
/* ------------------------------------------------------------ */
   static const Word16   BitClass63[5]    = { 44+CrcLen, 44, 46, 47+4, 12 };
   static const Word16   Weight63[5]      = { -8520, -9502, -7864, -6882, 0};

   static const Word16   InvBitClass63[4] = {21400, 23831, 22795, 22310};
   static const Word16   inv_b0p4_63      = 19785;

/* Definition of SID configuration (see Table C5b)              */
/* Note that the weight is in 16 bit representation and negated */
/* ------------------------------------------------------------ */
   static const Word16   BitClassSID[5] = { 30+CrcLen+4, 0, 0, 0, 0 };
   static const Word16   WeightSID[5]   = { -32768, 0, 0, 0, 0};

   static const Word16   inv_b0p4_SID   = 26887;

   Word16   i,j,rbits,r[5],abits,last,cbits,limit,NInfoBits;
   Word16   weight[5],inv_b0p4,inv_bit_class[4];
   Word32   L_tmp;

   /* Determine the number of available channel bits per frame (multiply by 0.03) */
   /* ----------------------------------------------------------------------------*/
   L_tmp = L_mult(ChannelBitrate,31458);
   cbits = (Word16) L_shr(L_tmp,21);

   if ((SpeechCodecMode & 0x3) == 0){
      for (i=0;i<5;i++) bit_class[i] = BitClass63[i];
      for (i=0;i<5;i++) weight[i] = Weight63[i];
      inv_b0p4 = inv_b0p4_63;
      for (i=0;i<4;i++) inv_bit_class[i] = InvBitClass63[i];
      NInfoBits = NInfoBits6_3;
   }
   if ((SpeechCodecMode & 0x3) == 1){
      for (i=0;i<5;i++) bit_class[i] = BitClass53[i];
      for (i=0;i<5;i++) weight[i] = Weight53[i];
      inv_b0p4 = inv_b0p4_53;
      for (i=0;i<4;i++) inv_bit_class[i] = InvBitClass53[i];
      NInfoBits = NInfoBits5_3;
   }
   if ((SpeechCodecMode & 0x3) == 2){
      for (i=0;i<5;i++) bit_class[i] = BitClassSID[i];
      for (i=0;i<5;i++) weight[i] = WeightSID[i];
      inv_b0p4 = inv_b0p4_SID;
      NInfoBits = NInfoBitsSID;
   }

   /* Determine the number of available redundancy bits per frame */
   /* ----------------------------------------------------------- */
   rbits = cbits - (NInfoBits+CrcLen+BCHLen);

   /* Check whether enough bits are available */
   /* --------------------------------------- */
   if (rbits < 0) {
      printf ("\n\nNot enough channel bitrate specified\n");
      printf ("The minimum channel bitrate is %.2f kbit/s \n",
            ((float)(BCHLen+CrcLen+NInfoBits)*1000./30.));
      exit(1);
   }

   if (rbits == 0) {
      /* No convolutional code specified */
      /* ------------------------------- */
      if (SpeechCodecMode < 2) {
         bit_class[3] = bit_class[3]-4;
      }else{
         bit_class[0] = bit_class[0]-4;
      }
      for (i=0;i<4;i++) {
         bit_class[4] = bit_class[4] + bit_class[i];
         bit_class[i] = 0;
      }
      for (i=0;i<5;i++){
         r[i]=12;
         for (j=0;j<3;j++) punctur[i][j] = PuncturTable[r[i]-12][j];
      }
   }else{

      /* Check whether too much bits have been specified */
      /* ----------------------------------------------- */
      limit = 2*(bit_class[0]+bit_class[1]+bit_class[2]+
                 bit_class[3]+bit_class[4]) + 4;
      if (rbits >= limit) {
	 if (rbits > limit) {
	    printf ("\n\nSpecified channel bitrate in bitrate configuration file too high\n");
	    printf ("The channel bitrate is limited to %.2f kbit/s \n",
                  (float)(3*(bit_class[0]+bit_class[1]+bit_class[2]+
                 bit_class[3]+bit_class[4])+BCHLen)*1000./30.);
         }
         for (i=0;i<4;i++)   r[i]=36;
         r[4]=12;
         for (i=0;i<5;i++){
            for (j=0;j<3;j++) punctur[i][j] = PuncturTable[r[i]-12][j];
         }
         bit_class[3] = bit_class[3] + bit_class[4];
         bit_class[4] = 0;
      }else{
         /* Initialise rates and puncturing tables */
         /* -------------------------------------- */
         for (i=0;i<5;i++)   r[i]=12;
         for (i=0;i<5;i++){
            for (j=0;j<3;j++) punctur[i][j] = PuncturTable[r[i]-12][j];
         }

         /* Determine rate for the most sensitive bit class */
         /* ----------------------------------------------- */
         rbits = rbits - 4; /* tail bits have to be transmitted */
         if (rbits < 0) rbits=0;
         if ((rbits <= bit_class[0]+8) || (SpeechCodecMode == 2)) {
            /* Allocate all redundancy to class 0 */
            if (SpeechCodecMode < 2) {
               bit_class[3] = bit_class[3]-4;
            }else{
               bit_class[0] = bit_class[0]-4;
            }
	    r[0] = GetRate(rbits,(Word16)(-32768),bit_class[0]+4,inv_b0p4,0);
            if (r[0] == 12) {
               for (i=0;i<4;i++) {
                  bit_class[4] = bit_class[4] + bit_class[i];
                  bit_class[i] = 0;
               }
            }else{
               bit_class[0] = bit_class[0] + 4;
            }
            for (j=0;j<3;j++) punctur[0][j] = PuncturTable[r[0]-12][j];
         }else{
	    r[0] = GetRate(rbits,weight[0],bit_class[0],inv_bit_class[0],1);
            if (r[0] < 24) r[0] = 24;
            if (r[0] > 36) r[0] = 36;
            for (j=0;j<3;j++) punctur[0][j] = PuncturTable[r[0]-12][j];
            bit_class[3] = bit_class[3] - 4;
            bit_class[0] = bit_class[0] + 4;
            for (i=1;i<4;i++) {
               abits = bitrate_test (bit_class,punctur,i-1);
               if (abits < cbits) {
		  r[i] = GetRate(rbits,weight[i],bit_class[i],inv_bit_class[i],1);
                  if (r[i] <  18) r[i] =  18;
                  if (r[i] > 36) r[i] = 36;
                  if (r[i] > r[0]) r[i] = r[0];
                  for (j=0;j<3;j++) punctur[i][j] = PuncturTable[r[i]-12][j];
                  bit_class[i-1] = bit_class[i-1] - 4;
                  bit_class[i]   = bit_class[i]   + 4;
               }else {
                  break;
               }
            }
         }
      }
   }

   /* Determine which class is the last protected one */
   /* ----------------------------------------------- */
   last=0;
   for (i=1;i<5;i++){
      if (r[i] > 12) last=i;
   }

   /* Determine channel bitrate with these settings */
   /* --------------------------------------------- */
   abits = bitrate_test (bit_class,punctur,last);
   if (abits > cbits) {
      /* too many bits are used: shift bits from high to low protection */
      /* -------------------------------------------------------------- */
      if (SpeechCodecMode < 2){
         bitrate_adapt_high (bit_class,punctur,cbits,r,last);
      }else{
         r[0]--;
         for (j=0;j<3;j++) punctur[0][j] = PuncturTable[r[0]-12][j];
         if (r[0] == 12) bit_class[0]=bit_class[0]-4;
      }
   }
   if ((abits < cbits) && (r[0]>12) && (rbits < limit) && (SpeechCodecMode < 2)) {
      /* too few bits are used:: shift bits from low to high protection */
      /* -------------------------------------------------------------- */
      bitrate_adapt_low (bit_class,punctur,cbits,r,last);
   }
   return;
}

/*_________________________________________________________________________________
 |                                                                                 |
 | Determine rate for one bit class                                                |
 |               (section 2.5.1)                                                   |
 |_________________________________________________________________________________|
*/
static Word16 GetRate (Word16 rbits, Word16 weight, Word16 bit_class, Word16 inv_bit_class, Word16 rnd)
{
         
   Word16 tmp;
   Word32 L_tmp1, L_tmp2;

   /* Calculate 12*(rbits*weight+bit_class) */
   /* ------------------------------------- */
   L_tmp1 = - L_mult(rbits,weight);
   L_tmp2 = - L_mult(bit_class,(Word16)(-32768));
   L_tmp1 = L_add(L_tmp1,L_tmp2);

   /* Multiply by 12 */
   /* -------------- */
   L_tmp2 = L_shl(L_tmp1,3);
   L_tmp1 = L_shl(L_tmp1,2);
   L_tmp1 = L_add(L_tmp1,L_tmp2);
   tmp   = (Word16) L_shr(L_tmp1,12);

   /* Multiply by inv_bit_class and round */
   /* ----------------------------------- */
   L_tmp1 = L_mult(tmp,inv_bit_class);
   tmp    = (Word16) L_shr(L_tmp1,24);
   tmp    = add(tmp,rnd);
   tmp    = shr (tmp,1);
   return(tmp);
}

/*_________________________________________________________________________________
 |                                                                                 |
 | Adaptation of bit allocation to meet channel bitrate requirement                |
 |               (section 2.5.1)                                                   |
 |_________________________________________________________________________________|
*/
static void bitrate_adapt_low (Word16 bit_class[],Word16 punctur[5][3],
                               Word16 cbits, Word16 r[], Word16 last)
{
   Word16 abits,src,dest;

   src  = last+1;
   dest = last;
   for (;;) {
      if (bit_class[src] == 0) {
         if (src == last) return;
         src = last;
         dest = last-1;
         if (r[src] == r[dest]) return;
      }
      /* shift one bit from bit_class[src] to bit_class[dest] */
      /* ---------------------------------------------------- */
      bit_class[src]--;
      bit_class[dest]++;
      abits = bitrate_test (bit_class,punctur,last);
      if (abits >= cbits) break;      
   }

   if (abits > cbits) bitrate_adapt_high (bit_class,punctur,cbits,r,last);

   return;
}

/*_________________________________________________________________________________
 |                                                                                 |
 | Adaptation of bit allocation to meet channel bitrate requirement                |
 |               (section 2.5.1)                                                   |
 |_________________________________________________________________________________|
*/
static void bitrate_adapt_high (Word16 bit_class[],Word16 punctur[5][3],
                                Word16 cbits, Word16 r[], Word16 last)
{
   Word16 abits,i;

   if (r[3] > 21) {
      /* The number of bits in the unprotected class may not be increased */
      /* ---------------------------------------------------------------- */
      for (;;) {
         for (i=2;i>0;i--){
            /* shift one bit of bit_class[i] to bit_class[i+1] */
            /* ----------------------------------------------- */
            if (bit_class[i] == 0) return;
            bit_class[i]--;
            bit_class[i+1]++;
            abits = bitrate_test (bit_class,punctur,last);
            if (abits <= cbits) break;      
         }
         if (abits <= cbits) break;      
      }
   }else{
      for (;;) {
         /* shift one bit of bit_class[last] to bit_class[last+1] */
         /* ----------------------------------------------------- */
         if (bit_class[last] == 0) {
            printf ("\n\n Error in bitrate_adapt_high\n\n");
            exit(1);
         }
         bit_class[last]--;
         bit_class[last+1]++;
         abits = bitrate_test (bit_class,punctur,last);
         if (abits <= cbits) break;
      }
   }
   return;
}


/*_______________________________________________________________________
 |                                                                      |
 | Determination of required channel bitrate                            |
 |               (section 2.5.1)                                        |
 |______________________________________________________________________|
*/
static bitrate_test( Word16 nbit_class[], Word16 punctur[5][3], Word16 last )
{
    Word16 i, j, k, ptr0, cnt;
    static const Word16 mask[16] = { 0X0001, 0X0002, 0X0004, 0X0008,
                                   0X0010, 0X0020, 0X0040, 0X0080,
                                   0X0100, 0X0200, 0X0400, 0X0800,
                                   0X1000, 0X2000, 0X4000, -32768 };

    /* Encode the protected classes: */
    /* ----------------------------- */
    ptr0 = 0;
    cnt  = 0;
    for( k = 0; k <= last; ++k ) {
        for( i = 0; i < nbit_class[k]; ++i ) {
            for( j = 0; j < 3; ++j ) {
                if(( punctur[k][j] & mask[cnt] ) != 0 ) ptr0++;
            }
            cnt = ++cnt % 12;
        }
    }

    /* Determine channel bitrate */
    /* ------------------------- */
    for (i=last+1;i<5;i++) ptr0 = ptr0 + nbit_class[i];

    ptr0 += BCHLen;

    return(ptr0);

}


static void hamming_distance( Word16 chan_bit_np[], Word16 metric[], Word16 temp[] )
{
#define SCALE 2
    static const Word16 code_word[16][3] = { 
        {  16384/SCALE,  16384/SCALE,  16384/SCALE }, 
        { -16384/SCALE, -16384/SCALE, -16384/SCALE },
        { -16384/SCALE,  16384/SCALE, -16384/SCALE }, 
        {  16384/SCALE, -16384/SCALE,  16384/SCALE }, 
        {  16384/SCALE, -16384/SCALE, -16384/SCALE }, 
        { -16384/SCALE,  16384/SCALE,  16384/SCALE },
        { -16384/SCALE, -16384/SCALE,  16384/SCALE }, 
        {  16384/SCALE,  16384/SCALE, -16384/SCALE }, 
        {  16384/SCALE, -16384/SCALE,  16384/SCALE },
        { -16384/SCALE,  16384/SCALE, -16384/SCALE },
        { -16384/SCALE, -16384/SCALE, -16384/SCALE }, 
        {  16384/SCALE,  16384/SCALE,  16384/SCALE }, 
        {  16384/SCALE,  16384/SCALE, -16384/SCALE }, 
        { -16384/SCALE, -16384/SCALE,  16384/SCALE },
        { -16384/SCALE,  16384/SCALE,  16384/SCALE },
        {  16384/SCALE, -16384/SCALE, -16384/SCALE }};
    Word16 ham_dist, i, j;
    Word32 acc;

    for( i = 0, j = 0; i < 8; ++i ) {
        acc = L_mac(   0, code_word[j][0], chan_bit_np[0] );
        acc = L_mac( acc, code_word[j][1], chan_bit_np[1] );
        acc = L_mac( acc, code_word[j][2], chan_bit_np[2] );
        ham_dist = extract_h( acc );
        temp[j] = add( metric[i], ham_dist );
        temp[j+17] = add( metric[i+8], ham_dist );
        ++j;
        acc = L_mac(   0, code_word[j][0], chan_bit_np[0] );
        acc = L_mac( acc, code_word[j][1], chan_bit_np[1] );
        acc = L_mac( acc, code_word[j][2], chan_bit_np[2] );
        ham_dist = extract_h( acc );
        temp[j] = add( metric[i], ham_dist );
        temp[j+15] = add( metric[i+8], ham_dist );
        ++j;
    }
}

static void metric_update( Word16 chan_bit_np[], Word16 metric[], Word16 *surv_mem )
{
    Word16 temp[32], j;

    hamming_distance( chan_bit_np, metric, temp );
    for( j = 0; j < 16; ++j ) {
        if( sub( temp[j], temp[j+16] ) >= 0 ) {
            metric[j] = temp[j];
	    if (metric[j] > 32760) printf("metric problem 1 %d\n",metric[j]);
	    if (metric[j] < -32760) printf("metric problem 1 %d\n",metric[j]);
            clrbit( *surv_mem, j );
        } else {
            metric[j] = temp[j+16];
            setbit( *surv_mem, j );
	    if (metric[j] > 32760) printf("metric problem 2 %d\n",metric[j]);
	    if (metric[j] < -32760) printf("metric problem 2 %d\n",metric[j]);
        }
    }
}

static Word16 metric_difference( Word16 chan_bit_np[], Word16 metric[], Word16 adr )
{
    Word16 delta, temp[32], j;

    hamming_distance( chan_bit_np, metric, temp );
    for( j = 0; j < 16; ++j )
        metric[j] = ( sub( temp[j], temp[j+16] ) >= 0 ) ? temp[j] : temp[j+16];

    /* Compute difference metric: */
    /* -------------------------- */
    delta = abs_s( sub( temp[adr], temp[adr+16] ));

    /* Scale difference metric: */
    /* ------------------------ */
    return( shr( delta, 1 ));
}


static void metric_update_delta( Word16 chan_bit_np[], Word16 metric[],
    Word16 *surv_mem, Word16 delta_soft[] )
{
    Word16 temp[32], j;

    hamming_distance( chan_bit_np, metric, temp );
    for( j = 0; j < 16; ++j ) {

        /* Compute difference metric: */
        /* -------------------------- */
        delta_soft[j] = sub( temp[j], temp[j+16] );

        if( delta_soft[j] >= 0 ) {
            metric[j] = temp[j];
            clrbit( *surv_mem, j );
        } else {
            metric[j] = temp[j+16];
            setbit( *surv_mem, j );
        }
    }
}




static void viterbi( Word16 nbit_class[], Word16 punctur[5][3], 
                     Word16 chan_bit_r[], Word16 info_bit_r[], 
                     Word16 info_bit_r_soft[])
{
    Word16 chan_bit_np[1500][3], surv[200], surv_mem[200];
    Word16 ninfo_conv, metric[16];
    Word16 ptr0, ptr1, adr, cnt12, i, j, k, last, n0;
    static const Word16 mask[16] = { 0X0001, 0X0002, 0X0004, 0X0008,
                                   0X0010, 0X0020, 0X0040, 0X0080,
                                   0X0100, 0X0200, 0X0400, 0X0800,
                                   0X1000, 0X2000, 0X4000, -32768 };


    if (punctur[0][0] != 0) {

       /* Check which class is the last protected one */
       last=0;
       for (i=1;i<5;i++){
          if (punctur[i][0] != 0) last=i;
       }
       for (ninfo_conv=0,i=0;i<=last;i++)  ninfo_conv += nbit_class[i];
       for (n0=0,i=last+1;i<5;i++)  n0 += nbit_class[i];
   
   
       /* Initialise metrics: */
       /* ------------------- */
       metric[0] = 0;
       for( i = 1; i < 16; ++i )
           metric[i] = -5000;
   
       /* --------------------------------------------------- */
       /* | Process trellis segments for protected classes  | */
       /* --------------------------------------------------- */
       ptr1  = BCHLen;
       cnt12 = -1;
       ptr0  = 0;
       for( k = 0; k <= last; ++k ) {
           for( i = 0; i < nbit_class[k]; ++i ) {
   
               /* Include zero values into punctured channel bit stream  */
               /* -------------------------------------------------------*/
               cnt12 = ++cnt12 % 12;
               for( j = 0; j < 3; ++j ) {
                   if(( punctur[k][j] & mask[cnt12] ) != 0 ) {
                       chan_bit_np[ptr0][j] = shl( chan_bit_r[ptr1++], 1 );
                   } else {
                       chan_bit_np[ptr0][j] = 0;
                   }
               }

               /* Update metrics without storage of metric differences: */
               /* ----------------------------------------------------- */
               metric_update( chan_bit_np[ptr0], metric, &surv_mem[ptr0] );
               ++ptr0;
           }
       }

       /* Hard decision decoding of information bits: */
       /* ------------------------------------------- */
       adr = 0;
       for( i = ninfo_conv - 1; i >= 0; --i ) {
           surv[i] = adr & 1;
           if(( surv_mem[i] & mask[adr] ) != 0 ) adr += 16;
           adr = shr( adr, 1 );
       }
   
       /* --------------------------------------------------------------------*/
       /* | Determine soft decision value for each protected information bit |*/
       /* --------------------------------------------------------------------*/
       /* this algorithm is not included in this version */
   


       /* Determine output from hard decision values: */
       /* ------------------------------------------- */
       ptr0 = ptr1;
       for( i = 0; i < ninfo_conv - 4; ++i ) info_bit_r[i] = surv[i];
       for( i = ninfo_conv - 4; i < ninfo_conv + n0 - 4; ++i ) {
          if (chan_bit_r[ptr1++] < 0)
             info_bit_r[i] = 1;
          else
             info_bit_r[i] = 0;
       }

       /* Copy soft decision values of unprotected class: */
       /* ----------------------------------------------- */
           /* insert here the values from soft decision evaluation */
       for( i = 0; i < ninfo_conv +n0- 4; ++i ) info_bit_r_soft[i] = 0x7f;

    }else{

       /* no convolutional code has been specified */
       ptr0 = BCHLen;
       for( i = 0; i < nbit_class[4]; ++i ) {
           info_bit_r_soft[i] = 0x7f;
           if (chan_bit_r[ptr0++] < 0)
              info_bit_r[i] = 1;
           else
              info_bit_r[i] = 0;
       }
    }
}




static void ChannelDecoder(Word16 chan_bit_r[], char InfoBitStream_r[])
{
    static char first=0;
    static Word16 LastMode_r={-1};
    static Word16 bit_class[5],punctur[5][3];
    Word16   info_bit_r[210], Bfi,Efi,Fii,bad_frame, bad_frame_soft[2];
    Word16   info_bit_r_soft[210],IndexError[3];
    Word16   mode_r, err_conf, i, ModeAccept, G723_mode_r, CC_mode_r;
    Word16   SysChannelBitrate_r;
    char   CInfoBitStream_r[25];


    /* Decode configuration bits: */
    /* -------------------------- */
    ConfigDecode (chan_bit_r, &mode_r, &err_conf);
    ModeChangeAccept (mode_r, err_conf, &ModeAccept);


    if (ModeAccept == 1) {
       G723_mode_r = mode_r & 0x3;
       CC_mode_r   = mode_r >> 2;

       if (mode_r != LastMode_r) {
          /* Re-Initialise Channel Codec */
          /* --------------------------- */
          DecodeChannelBitrate(CC_mode_r,G723_mode_r,&SysChannelBitrate_r);
          ChannelCodecInit(G723_mode_r,SysChannelBitrate_r,bit_class,punctur);
          LastMode_r = mode_r;
       }

       /* Viterbi decoder with soft input and output: */
       /* ------------------------------------------- */
       viterbi(bit_class, punctur, chan_bit_r, info_bit_r, info_bit_r_soft);


       /* Consider low residual error probability at the end of the trellis */
       /* ----------------------------------------------------------------- */
       EndTrellisReorderDec (info_bit_r, info_bit_r_soft, bit_class, punctur, G723_mode_r);

       /* Determine badframe flag from CRC code */
       /* ------------------------------------- */
       CrcDecoder(info_bit_r,&bad_frame, G723_mode_r);

       /* Improve badframe decision by soft output evaluation */
       /* --------------------------------------------------- */
       SoftBFI (bit_class,punctur,info_bit_r_soft, bad_frame_soft, bad_frame, G723_mode_r);

       /* Reconstruct bitstream */
       /* --------------------- */
       bit_deallocation(info_bit_r, CInfoBitStream_r, G723_mode_r);

       /* Adapt bitstream to G.723.1 decoder */
       /* ---------------------------------- */
       speech_decoder_adaptation(CInfoBitStream_r,InfoBitStream_r,IndexError);

    }else{
       bad_frame         = 1;
       bad_frame_soft[0] = 1;
       bad_frame_soft[1] = 1;
    }

    /* Set concealment flags */
    /* --------------------- */
    Bfi = 0;
    if ((bad_frame | bad_frame_soft[0]) == 1) Bfi = 1;

    Efi = bad_frame_soft[1];

    Fii = 0;
    if (IndexError[0]+IndexError[1]+IndexError[2] > 0) Fii=1;

    /* Add concealment information as last byte to the compressed bitstream */
    /* -------------------------------------------------------------------- */
    if ((InfoBitStream_r[0] & 0x3) == 0) i=24;
    if ((InfoBitStream_r[0] & 0x3) == 1) i=20;
    if ((InfoBitStream_r[0] & 0x3) == 2) i= 4;
    InfoBitStream_r[i] = 0;
    InfoBitStream_r[i] = InfoBitStream_r[i] |  Bfi;
    InfoBitStream_r[i] = InfoBitStream_r[i] | (Efi << 1);
    InfoBitStream_r[i] = InfoBitStream_r[i] | (Fii << 2);

    return;
}


static void ModeChangeAccept (Word16 mode_r, Word16 err_conf, Word16 *ModeAccept)
{
   static Word16 mode_r_old = {0};
   Word16   AcceptThreshold;

   AcceptThreshold=1000;

   if (mode_r == mode_r_old){
      *ModeAccept = 1;
   }else{
      /* Accept the mode change if the decoding was sufficiently reliable */
      /* ---------------------------------------------------------------- */
      if (err_conf > AcceptThreshold){
         /* Check whether G.723.1 mode is valid */
         if ( (mode_r & 0x3) == 3 ) {
            *ModeAccept = 0;
         }else{
            *ModeAccept = 1;
            mode_r_old = mode_r;
         }
      }else{
         *ModeAccept = 0;
      }
   }
   return;
}


static void ConfigDecode (Word16 chan_bit_r[], Word16 *mode_r, Word16 *err_conf)
{

/* Definition of BCH Code for configuration bits (see section 2.5.4) */
/* ----------------------------------------------------------------- */
static const Word16 ConfigGenMatrix[NumBCHWords][BCHLen] = {
{ 127, 127, 127, 127, 127, 127, 127, 127, 127, 127, 127, 127, 127},
{-127,-127,-127, 127,-127,-127, 127, 127,-127, 127,-127, 127, 127},
{ 127,-127,-127,-127, 127,-127,-127, 127, 127,-127, 127,-127, 127},
{-127, 127, 127,-127,-127, 127,-127, 127,-127,-127,-127,-127, 127},
{ 127, 127,-127,-127,-127, 127,-127,-127, 127, 127,-127, 127,-127},
{-127,-127, 127,-127, 127,-127,-127,-127,-127, 127, 127, 127,-127},
{ 127,-127, 127, 127,-127,-127, 127,-127, 127,-127,-127,-127,-127},
{-127, 127,-127, 127, 127, 127, 127,-127,-127,-127, 127,-127,-127}};

   Word32   dist_max, dist[NumBCHWords], dist_max2, i, j, tmp;
   static Word16 first={0}; /* for error analysis only */

   /* Determine distance between received codeword and stored codewords */
   for (i=0;i<NumBCHWords;i++){
      dist[i] = 0;
      for (j=0;j<BCHLen;j++) dist[i] += chan_bit_r[j] * ConfigGenMatrix[i][j];
   }

   /* Perform maximum likelihood decoding */
   dist_max = -2147483647;
   for (i=0;i<NumBCHWords;i++){
      if (dist[i] > dist_max) {
         *mode_r = i;
         dist_max = dist[i];         
      }
   }

   /* Determine decoding reliability */
   dist[*mode_r] = -2147483647;
   dist_max2 = -2147483647;
   for (i=0;i<NumBCHWords;i++){
      if (dist[i] > dist_max2) {
         dist_max2 = dist[i];
      }
   }

   tmp = (dist_max-dist_max2) >> 3;
   if (tmp > 32767){
      *err_conf = 32767;
   }else{
      *err_conf = tmp;
   }

   return;
}


static void  Init_DecoderState (void)
{
   Word16 i;

   DecStat.Ecount=0 ;
   DecStat.InterGain=0 ;
   DecStat.InterIndx=0 ;
   DecStat.Rseed=0 ;
   DecStat.Park=0  ;
   DecStat.Gain=0  ;
   for(i=0;i<LpcOrder;i++) DecStat.PrevLsp[i]=0 ;
   for(i=0;i<PitchMax;i++) DecStat.PrevExc[i]=0 ;
   for(i=0;i<LpcOrder;i++) DecStat.SyntIirDl[i]=0 ;
   for(i=0;i<LpcOrder;i++) DecStat.PostFirDl[i]=0 ;
   for(i=0;i<LpcOrder;i++) DecStat.PostIirDl[i]=0 ;

   return;
}


static void bit_deallocation(Word16 info_bit_r[], char CInfoBitStream_r[], 
                             Word16 G723_mode_r )
{

static const Word16   BitOrder63[NBits6_3] = 
       {175, 180, 189, 190, 191, 192 , -1 , -1,
         98 , 73, 107, 154, 167, 168, 169, 170,
         30 , 17 , 16 , 31 , 48 , 55 , 49 , 71,
          6  , 4  , 0  , 2 , 11 , 26 , 10 , 14,
          5  , 1  , 3 , 12 , 27 , 24 , 60  , 8,
         44 , 66 , 62 , 82 , 25 , 61  , 9  , 7,
         45 , 67 , 63 , 83 , 78 , 36 , 50 , 40,
         46 , 68 , 64 , 84 , 79 , 37 , 51 , 41,
         47 , 69 , 65 , 85 , 80 , 38 , 52 , 42,
         56 , 99, 159, 185 , 81 , 39 , 53 , 43,
        161, 187 , 20,  57 ,100, 160, 186 , 19,
         22,  59 ,102, 162, 188 , 21,  58 ,101,
         35 , 54 , 70 , 72, 179, 178, 177, 176,
         13 , 15 , 23 , 28 , 29 , 32 , 33 , 34,
         128, 132, 146, 155, 163, 171, 181, 18,
         76 , 86 , 90 , 94 ,103, 108, 112, 116,
        129, 133, 147, 156, 164, 172, 182 , 74,
        183 , 87 , 91 , 95, 104, 109, 113, 117,
        114, 118, 130, 134, 148, 157, 165, 173,
        184 , 75 , 77 , 88 , 92 , 96, 105, 110,
        115, 119, 131, 135, 149, 158, 166, 174,
        136, 124, 120 , 89 , 93 , 97, 106, 111,
        151, 141, 137, 125, 121, 144, 150, 140,
        127, 123, 145, 152, 142, 138, 126, 122,
         -1 , -1 , -1 , -1 , -1, 153, 143, 139};


static const Word16   BitOrder53[NBits5_3] = 
{152, 153, 158, 159, 160, 161 , -1 , -1,
  69 , 64 , 70 , 91, 145, 140, 147, 146 ,
  24 , 15 , 14 , 25 , 46 , 50 , 47,  63,
   4  , 6  , 0  , 2 , 11 , 18 , 10,  13 ,
   7  , 1  , 3 , 12 , 19 , 16 , 48,   8,
  42 , 59 , 55 , 65 , 17 , 49  , 9,   5   ,
  43 , 60 , 56 , 66 , 26 , 30 , 34,  38  ,
  44 , 61 , 57 , 67 , 27 , 31 , 35,  39  ,
  45 , 62 , 58 , 68 , 28 , 32 , 36,  40  ,
  51 , 87 ,141, 154 , 29 , 33 , 37,  41  ,
 143, 156 , 21,  52 , 88 ,142, 155,  20 ,
  23,  54 , 90 ,144, 157 , 22,  53,  89  ,
 100, 128 , 96, 104, 151, 150, 149, 148,
 132, 112, 116, 136, 108, 120, 124,  92 ,
 109, 121, 125 , 93, 101, 129 , 97, 105,
 102, 130 , 98, 106, 133, 113, 117, 137 ,
 134, 114, 118, 138, 110, 122, 126,  94 ,
 111, 123, 127 , 95, 103, 131 , 99, 107,
  83 , 79 , 75 , 71, 135, 115, 119, 139 ,
  85 , 81 , 77 , 73 , 84 , 80 , 76,  72,
  -1 , -1 , -1 , -1 , 86 , 82 , 78,  74};


static const Word16   BitOrderSID[NBitsSID] = 
       {  21, 22, 23, 24, 25, 26, -1, -1,
          13, 14, 15, 16, 17, 18, 20, 19,
           5,  6,  7,  8,  9, 10, 11, 12,
           0,  1,  2, 29, 28, 27,  3,  4};


    Word16        i, pw, pb, j, info_bit_hd[NInfoBits6_3];
    Word16        BitOrder[NBits6_3], EndLoop, crc_window, nbit_all;


   if ((G723_mode_r & 3) == 0){
      /* Rate 6.3 */
      EndLoop = NBits6_3;
      for (i=0;i<EndLoop;i++) BitOrder[i] = BitOrder63[i];
      crc_window = CrcWindow63;
      nbit_all = NInfoBits6_3+CrcLen+4;
   }
   if ((G723_mode_r & 3) == 1){
      /* Rate 5.3 */
      EndLoop = NBits5_3;
      for (i=0;i<EndLoop;i++) BitOrder[i] = BitOrder53[i];
      crc_window = CrcWindow53;
      nbit_all = NInfoBits5_3+CrcLen+4;
   }
   if( (G723_mode_r & 0x3) == 2) {   
     /* Silence */
      EndLoop = NBitsSID;
      for (i=0;i<EndLoop;i++) BitOrder[i] = BitOrderSID[i];
      crc_window = CrcWindowSID;
      nbit_all = NInfoBitsSID+CrcLen+4;
   }
   if( (G723_mode_r & 0x3) == 3) {   
     /* not used */
     exit;
   }


   for (i = 0; i < crc_window; ++i) {  /* without CRC */
      info_bit_hd[i] = info_bit_r[i];
   }
   for (i=crc_window+CrcLen; i<nbit_all-4; ++i) {  /* without Tail */
      info_bit_hd[i-CrcLen] = info_bit_r[i];
   }

    for( i=0; i < 25; i++ )  CInfoBitStream_r[i]=0;

    for( i=0; i < EndLoop; i++ ) {
       pw = i/8;
       pb = i - pw*8;
       j  = (pw*8)+7-pb;
       if (BitOrder[j] != (-1)){
          if (info_bit_hd[BitOrder[j]] == 1) setbit( CInfoBitStream_r[pw], pb );
       }
    }       
    CInfoBitStream_r[0] = CInfoBitStream_r[0] | G723_mode_r;
    return;
}




static void   CrcDecoder(Word16 info_bit_r[], Word16 *bad_frame, Word16 G723_mode_r)
{

   Word16        crc_window;
   Word16        CrcCarry, ShiftReg, i;

   if ((G723_mode_r & 0x3) == 0){
      /* Rate 6.3 */
      crc_window = CrcWindow63;
   }
   if ((G723_mode_r & 0x3) == 1){
      /* Rate 5.3 */
      crc_window = CrcWindow53;
   }
   if( (G723_mode_r & 0x3) == 2) {   
     /* Silence */
      crc_window = CrcWindowSID;
   }
   if( (G723_mode_r & 0x3) == 3) {   
     /* not used */
     exit;
   }

   ShiftReg = 0;
   for (i = 0; i < crc_window; ++i) {
     CrcCarry = (info_bit_r[i] ^ ShiftReg) & 1;
     ShiftReg = shr(ShiftReg, 1);
     if (CrcCarry != 0) ShiftReg = ShiftReg ^ CrcGen;
   }


   /* determine CRC-badframe */
   /* ---------------------- */
   *bad_frame = 0;
   for (i=CrcLen-1;i>=0;i--){
      if (info_bit_r[crc_window + i] != (ShiftReg & (Word16) 1)) *bad_frame=1;
      ShiftReg = shr(ShiftReg, 1);
   }
   return;
}




static void LSP_Reorder_decoder (LINEDEF *Line)
{

/* Band0Tb8 */
static	Word16 tabd2_10[LspCbSize] = {
	 140,  185,  135,  137,  169,  168,  164,  166, 
	 187,  192,  238,  191,  180,  141,  177,  179, 
	 181,  183,  195,  184,  172,  173,  165,  161, 
	 194,  197,  196,  198,  188,  186,  189,  190, 
	   7,   43,   16,   17,    5,   40,    8,    9, 
	  20,   19,   22,   21,   42,  130,  131,  240, 
	 246,  249,  250,  251,  182,  239,  241,  242, 
	  23,   24,  247,  248,  253,  254,  252,  255, 
	 201,  143,  144,  136,  115,  114,  203,  202, 
	 139,  158,   28,  157,   97,   98,  228,  121, 
	 211,  171,    0,  174,  227,   27,  199,  200, 
	  87,  210,   25,    1,  219,  151,  162,  163, 
	 146,  142,  145,  170,  154,  138,  176,  178, 
	 243,  244,  236,  237,  105,  153,  155,  156, 
	  11,  175,  160,  159,   14,  132,  104,  103, 
	  15,  106,   18,  245,  134,  133,  193,  129, 
	  68,   67,  110,   65,   93,   92,   72,   71, 
	  90,   89,   88,   55,  225,   85,  224,  223, 
	  61,   60,   59,   50,   66,  234,  233,  230, 
	 205,  204,   51,   47,  231,   58,  147,   44, 
	  57,   69,  207,   56,  209,   53,   12,  208, 
	  26,   32,   29,  229,   33,  206,   31,   13, 
	  63,   62,   34,   37,   39,   38,   35,   45, 
	  36,   10,   41,    6,   46,    4,    2,    3, 
	 100,  102,  128,  101,   96,   95,  126,  222, 
	  74,   94,  220,   73,   77,   78,   75,   76, 
	 213,  221,  116,  122,   80,   82,  235,  216, 
	  86,   91,   54,   70,   83,   84,   81,  226, 
	 119,  120,  117,  118,  149,  150,  167,  152, 
	 214,  215,   30,  113,  217,  232,  212,  218, 
	 111,  112,  127,   99,   64,  109,   52,  108, 
	 148,  107,  123,  124,   49,   48,   79,  125
	};

/* Band1Tb8 */
static	Word16 tabd1_10[LspCbSize] = {
	 250,  249,  255,  254,  248,  251,  253,  252, 
	 150,  238,  243,  242,  226,  225,  227,  213, 
	 237,  234,  247,  246,  148,  147,  146,  145, 
	 240,  230,  244,  232,  239,  149,  212,  231, 
	 137,  221,  224,  223,  135,  134,  136,  141, 
	 220,  217,  222,  158,  153,  102,  105,  104, 
	 144,  143,  218,  219,  215,  157,  216,  152, 
	 151,   98,   75,   74,  155,  156,  103,  101, 
	 245,  241,  236,  235,   66,   65,   68,   67, 
	  70,   69,   64,   63,  208,  207,   60,  209, 
	 233,  140,  127,  126,  229,  228,  125,  139, 
	 131,  117,  123,  122,  133,  142,  124,  121, 
	 138,  214,  108,  106,  211,  210,  111,  109, 
	 120,  119,  113,  112,  206,  107,   18,  205, 
	  99,   76,   73,  160,  132,  130,  118,  129, 
	  97,   94,    0,   72,  110,   96,   95,   33, 
	 115,  114,   10,  116,   83,   81,   11,    9, 
	   8,   78,  128,   71,   12,   79,  163,  159, 
	  23,   25,   15,   24,   84,   87,    1,   82, 
	  22,   19,   62,   61,   20,   91,   21,   85, 
	  50,   45,   49,   48,  182,   51,  164,   47, 
	  40,   39,   77,  100,   35,   37,   36,  154, 
	 161,   43,   46,   38,  162,   41,   42,   44, 
	 165,   34,   80,   88,  166,   89,   93,   90, 
	 180,   28,  172,  171,   86,   27,   92,   26, 
	 181,    2,  184,  185,   29,   30,  168,  167, 
	  17,   16,  179,  177,    5,    3,    4,  178, 
	  13,  174,   14,  173,  195,  175,  199,  176, 
	 191,  192,  183,  187,  194,  188,  186,  169, 
	  52,   55,   53,   54,   57,   56,   58,   59, 
	 196,  197,  198,  193,    7,  200,    6,  203, 
	  32,   31,  202,  201,  170,  189,  204,  190
	};

/* Band0Tb8 */
static	Word16 tabd0_10[LspCbSize] = {
	 183,  196,  195,  179,  165,  180,  220,  221, 
	 203,  187,  206,  205,   36,   23,  198,  199, 
	 214,  178,  216,   26,  176,  177,  218,  219, 
	 141,  136,  142,  224,  155,  137,  229,  228, 
	 124,  117,  122,  123,  161,  169,  194,  170, 
	 138,  127,  134,  135,  201,  204,  149,  150, 
	 147,  168,  172,  171,  166,  167,  162,  164, 
	 133,  132,  130,  129,  121,  120,  128,  131, 
	 222,  223,  248,  249,  208,  209,   31,   32, 
	  35,   37,   24,   25,   28,   27,   30,   29, 
	 240,  253,  241,  242,  250,  252,  254,  255, 
	 234,  235,  236,  237,  246,  251,  239,  247, 
	 233,  230,  244,  238,  211,  245,  227,  210, 
	 152,  151,  215,  173,  154,  153,  207,  217, 
	 157,  158,  231,  232,  156,  226,  225,  243, 
	 140,  148,  146,  144,  145,  139,  126,  143, 
	  76,  185,   77,   41,   15,   43,  107,  110, 
	  58,   64,   65,   66,    6,  112,   60,   59, 
	  19,   78,   14,   21,   22,  163,   39,   40, 
	  92,   86,    1,  106,   93,  104,   18,  105, 
	 101,  102,   97,  100,   11,   12,   10,  103, 
	   9,    5,    3,    8,   69,   68,   67,   61, 
	 116,  114,   95,   87,   89,   88,   99,   98, 
	  94,   96,  108,  111,    7,    4,    2,  109, 
	  53,  192,  190,   55,   52,   54,   33,   34, 
	  48,   49,   56,   57,   42,   44,   63,   62, 
	 202,   72,  193,  174,  197,  212,  213,  175, 
	  13,  188,  186,  189,   38,  191,  182,  181, 
	  73,   47,   50,   51,   71,   46,   45,  184, 
	  85,   91,   17,   90,   80,   81,   74,   75, 
	  70,  200,   16,    0,  118,  119,  160,  159, 
	  83,  125,   79,  115,   20,   84,  113,   82
	};

Word16   lsp0,lsp1,lsp2;

/* Extract and reorder Band 2 */
   lsp2 = (Word16) (Line->LspId & (Word32) 0x000000FF);
   lsp2 = tabd2_10[lsp2];

/* Extract and reorder Band 1 */
   lsp1 = (Word16) ((Line->LspId >> 8) & (Word32) 0x000000FF);
   lsp1 = tabd1_10[lsp1];

/* Extract and reorder Band 0 */
   lsp0 = (Word16) ((Line->LspId >> 16) & (Word32) 0x000000FF);
   lsp0 = tabd0_10[lsp0];

/* Construct new Line->LspId */
   Line->LspId = (Word32) lsp0;
   Line->LspId = (Line->LspId << 8) + (Word32) lsp1;
   Line->LspId = (Line->LspId << 8) + (Word32) lsp2;

   return;
}




static void speech_decoder_adaptation (char CInfoBitStream_r[],
                                       char InfoBitStream_r[],Word16 IndexError[])
{
   Word16     FrType, i;
   LINEDEF  Line ;

   FrType = CInfoBitStream_r[0] & 3;

   if (FrType < 2) {
      Line = Line_Unpk_Channel( CInfoBitStream_r, IndexError) ;

      LSP_Reorder_decoder (&Line);

      /* This is a G.723.1 routine which uses the global "WrkRate"*/
      if (FrType == 0) WrkRate=Rate63;
      if (FrType == 1) WrkRate=Rate53;
      Line_Pack( &Line, InfoBitStream_r, (Word16) 1 ) ;   /* Ftyp = 1 */
      InfoBitStream_r[0] = InfoBitStream_r[0] & 0xFC;
      InfoBitStream_r[0] = InfoBitStream_r[0] | FrType;
   }else{
      for (i=0;i<4;i++) InfoBitStream_r[i] = CInfoBitStream_r[i];
      for (i=0;i<3;i++) IndexError[i] = 0;
   }
   return;
}



static LINEDEF  Line_Unpk_Channel( char *Vinp, Word16 IndexError[] )
{
        Word32    i,j;
	Word16	InfoBitStream_tmp[195] ;
	Word16    *Bsp = InfoBitStream_tmp ;
	Word16	FrType,error, Bound_AcGn ;
        LINEDEF Line ;
	Word32	Temp ;
        static Word16  MapIndexMamp[24]={0,1,2,3,4,5,6, 7, 8, 9,10,11,12,13,
                                       14,15,16,17,18,19,20,21,22,23};
        static Word16 Olp0_old={0},Olp1_old={0},AcGn_old={0},Mamp_old={0};
        static Word16 Tran_old={0};

        /* Clear index error counters */
        for (i=0;i<3;i++) IndexError[i]=0;

	/* Unpack the byte info to InfoBitStream vector */
	for ( i = 0 ; i < 192+3 ; i ++ )
		InfoBitStream_tmp[i] = ( Vinp[i>>3] >> (i & (Word16)0x0007) ) & (Word16)1 ;

	/* Decode the first two bits */
	FrType = (Word16)Ser2Par( &Bsp, 2 ) ;

	/* Decode the LspId */
	Line.LspId = Ser2Par( &Bsp, 24 ) ;

	if ( FrType == 2 ) {
		/* Decode the Noise Gain */

		return Line ;
	}

	/*
		Decode the common information to both rates
	*/

	/* Decode the adaptive codebook lags */
	Temp = Ser2Par( &Bsp, 7 ) ;
        if( Temp <= 123) {
           Line.Olp[0] = (Word16) Temp + (Word16)PitchMin ;
           Olp0_old = Line.Olp[0];
        }else {
           Line.Olp[0] = Olp0_old;
           IndexError[0] += 1;
        }

	Line.Sfs[1].AcLg = (Word16) Ser2Par( &Bsp, 2 ) ;

	Temp = Ser2Par( &Bsp, 7 ) ;
        if( Temp <= 123) {
           Line.Olp[1] = (Word16) Temp + (Word16)PitchMin ;
           Olp1_old = Line.Olp[1];
        }else {
           Line.Olp[1] = Olp1_old;
           IndexError[0] += 2;
        }

	Line.Sfs[3].AcLg = (Word16) Ser2Par( &Bsp, 2 ) ;

	Line.Sfs[0].AcLg = 1 ;
	Line.Sfs[2].AcLg = 1 ;

	/* Decode the 8 bit adaptive gains */
	for ( i = 0 ; i < SubFrames ; i ++ ) {

		Temp = Ser2Par( &Bsp, 8 ) ;

		Line.Sfs[i].Tran = 0 ;
                Bound_AcGn = NbFilt170 ;
		if ( (FrType == 0) && (Line.Olp[i>>1] < (SubFrLen-2) ) ) {
			Line.Sfs[i].Tran = (Word16)(Temp >> 7) ;
			Temp &= 0x0000007fL ;
                        Bound_AcGn = NbFilt085 ;
		}
                Line.Sfs[i].AcGn = (Word16) Temp;
                if((Line.Sfs[i].AcGn < Bound_AcGn)&&( (IndexError[0]&(1<<(i>>1)) ) == 0) ) {
                   AcGn_old = Line.Sfs[i].AcGn;
       		   Tran_old = Line.Sfs[i].Tran;
                }else{
       		   Line.Sfs[i].Tran = Tran_old ;
                   Line.Sfs[i].AcGn = AcGn_old;
                   IndexError[1]++;
                }
	}


	/* Decode the 5 bit fixed gains */
	for ( i = 0 ; i < SubFrames ; i ++ ) {
	   Line.Sfs[i].Mamp = (Word16) Ser2Par( &Bsp, 5 ) ;
           error = 1;
           for (j=0;j<NumOfGainLev;j++){
              if (Line.Sfs[i].Mamp == MapIndexMamp[j]){
                 Line.Sfs[i].Mamp = j;
                 error = 0;
              }
           }
           if (error == 0){
              Mamp_old = Line.Sfs[i].Mamp;
           }else{
              Line.Sfs[i].Mamp = Mamp_old;
              IndexError[2]++;
      	   }
	}


	/* Decode the grids */
	for ( i = 0 ; i < SubFrames ; i ++ )
		Line.Sfs[i].Grid = *Bsp ++ ;

	if ( FrType == 0 ) {

		/* Decode 13 bit combined position index */
		Temp = Ser2Par( &Bsp, 13 ) ;
		Line.Sfs[0].Ppos = ( Temp/90 ) / 9 ;
		Line.Sfs[1].Ppos = ( Temp/90 ) % 9 ;
		Line.Sfs[2].Ppos = ( Temp%90 ) / 9 ;
		Line.Sfs[3].Ppos = ( Temp%90 ) % 9 ;

		/* Decode all the pulse positions */
		Line.Sfs[0].Ppos = ( Line.Sfs[0].Ppos << 16 ) + Ser2Par( &Bsp, 16 ) ;
		Line.Sfs[1].Ppos = ( Line.Sfs[1].Ppos << 14 ) + Ser2Par( &Bsp, 14 ) ;
		Line.Sfs[2].Ppos = ( Line.Sfs[2].Ppos << 16 ) + Ser2Par( &Bsp, 16 ) ;
		Line.Sfs[3].Ppos = ( Line.Sfs[3].Ppos << 14 ) + Ser2Par( &Bsp, 14 ) ;
		
		/* Decode pulse amplitudes */
		Line.Sfs[0].Pamp = (Word16)Ser2Par( &Bsp, 6 ) ;
		Line.Sfs[1].Pamp = (Word16)Ser2Par( &Bsp, 5 ) ;
		Line.Sfs[2].Pamp = (Word16)Ser2Par( &Bsp, 6 ) ;
		Line.Sfs[3].Pamp = (Word16)Ser2Par( &Bsp, 5 ) ;
	}

	else {
		/* Decode the positions */
		for ( i = 0 ; i < SubFrames ; i ++ )
			Line.Sfs[i].Ppos = Ser2Par( &Bsp, 12 ) ;

		/* Decode the amplitudes */
		for ( i = 0 ; i < SubFrames ; i ++ )
			Line.Sfs[i].Pamp = (Word16)Ser2Par( &Bsp, 4 ) ;
	}
   return Line ;
}


static void SoftBFI (Word16 bit_class[], Word16 punctur[5][3], 
           Word16 info_bit_r_soft[], Word16 bad_frame_soft[], Word16 bad_frame,
	   Word16 G723_mode_r)
{
#define FrmHist 10

   static Word16   soft_history[20] = {127,127,127,127,127,127,127,127,127,127,
                                     127,127,127,127,127,127,127,127,127,127};
   static Word16   mean_class_last[5] = {127,127,127,127,127};
   Word16   crc_window,last,bclass[5];
   Word16   i,j,mean_hist,mean_class[5],ptr,cnt;
   double tmp;

   if ((G723_mode_r & 3) == 0){
      crc_window = CrcWindow63;
   }
   if ((G723_mode_r & 3) == 1){
      crc_window = CrcWindow53;
   }
   if( (G723_mode_r & 0x3) == 2) {   
      crc_window = CrcWindowSID;
   }

   bad_frame_soft[0]=0;
   bad_frame_soft[1]=0;

   if (punctur[0][0] != 0) {

      /* Check which class is the last protected one */
      last=0;
      for (i=1;i<5;i++){
         if (punctur[i][0] != 0) last=i;
      }
      for (i=0;i<=last-1;i++) bclass[i]=bit_class[i];
      bclass[last] = bit_class[last]-4;


      /* Determine mean soft value of last "FrmHist" frames */
      tmp = 0.;
      for (i=20-FrmHist;i<20;i++) tmp += (double) soft_history[i];
      mean_hist = (Word16) (tmp/(double)FrmHist);

      /* Determine mean soft value of actual frame and store */
      for (i=1;i<20;i++) soft_history[i-1] = soft_history[i];   
      tmp = 0.;
      for (i=0;i<crc_window;i++) tmp += (double) info_bit_r_soft[i];
      soft_history[19] = (Word16) (tmp/(double) crc_window)+0.5;

      /* Set soft badframe flag if actual soft value indicates unreliable decoding */
      if ((double)soft_history[19] < 0.17*(double)mean_hist) bad_frame_soft[0] = 1;
      if (soft_history[19] < 7) bad_frame_soft[0] = 1;

      /* Determine mean soft value for all bit classes */
      if (bclass[0]>crc_window+CrcLen){
         tmp = 0.;
         for (i=crc_window;i<bclass[0];i++) tmp += (double) info_bit_r_soft[i];
         mean_class[0] = tmp/(double)(bclass[0]-crc_window);
      }else{
         mean_class[0]=127;
      }
   
      ptr=bclass[0];
      for (i=1;i<=last;i++){
         if (bclass[i] > 0) {
            tmp = 0.;
            for (j=0;j<bclass[i];j++) tmp += (double) info_bit_r_soft[ptr++];
            mean_class[i] = tmp/(double)(bclass[i]);
         }else{
            mean_class[i]=127;
         }
      }      

      /* Set EFI flag if classes others than 0 are too unraliable */
      cnt=0;
      tmp = 0.35 * (double)soft_history[19];
      for (i=0;i<=last;i++) if (mean_class[i] < (Word16) tmp) cnt++;
      if (cnt >= 3) bad_frame_soft[1]=1;
         	
      cnt=0;
      for (i=0;i<=last;i++) if (mean_class[i] < 20) cnt++;
      if (cnt >= 2) bad_frame_soft[1]=1;
      cnt=0;
      for (i=0;i<=last;i++) if (mean_class[i] < 15) cnt++;
      if (cnt >= 1) bad_frame_soft[1]=1;

      cnt=0;
      for (i=1;i<=last;i++){
         tmp = 0.35 * (double)mean_class_last[i];
         if (mean_class[i] < (Word16) tmp) cnt++;
      }
      if (cnt >= 3) bad_frame_soft[1]=1;
      for (i=0;i<=last;i++) mean_class_last[i]=mean_class[i];

      if ((bad_frame==0)&&(bad_frame_soft[0]==1)&&(bad_frame_soft[1]==0)){
         bad_frame_soft[0]=0;
	 bad_frame_soft[1]=1;
      }

   }
   return;
}

static void  GetChannelBitrates(char *ConfigName)
{
    FILE   *fp;
    Word32   flag, i, tmp;
    char   InputLine[100];

    if((fp = fopen(ConfigName,"r"))==NULL){
       fprintf(stderr,"Can't open file '%s'\n",ConfigName);
       exit(1);
    }
    flag=0;
    while(flag == 0 ) {
      fgets( InputLine, 100, fp);
      if( InputLine[0] != '#') flag=1;
    }
    sscanf(InputLine, "%d", &tmp);
    ChannelBitrate_53[0] = (Word16) tmp;
    for (i=1;i<2;i++) {
      fgets( InputLine, 100, fp);
      sscanf(InputLine, "%d", &tmp);
      ChannelBitrate_53[i] = (Word16) tmp;
    }

    flag=0;
    while(flag == 0 ) {
      fgets( InputLine, 100, fp);
      if( InputLine[0] != '#') flag=1;
    }
    sscanf(InputLine, "%d", &tmp);
    ChannelBitrate_63[0] = (Word16) tmp;
    for (i=1;i<2;i++) {
      fgets( InputLine, 100, fp);
      sscanf(InputLine, "%d", &tmp);
      ChannelBitrate_63[i] = (Word16) tmp;
    }

    flag=0;
    while(flag == 0 ) {
      fgets( InputLine, 100, fp);
      if( InputLine[0] != '#') flag=1;
    }
    sscanf(InputLine, "%d", &tmp);
    ChannelBitrate_SID[0] = (Word16) tmp;
    for (i=1;i<2;i++) {
      fgets( InputLine, 100, fp);
      sscanf(InputLine, "%d", &tmp);
      ChannelBitrate_SID[i] = (Word16) tmp;
    }

    fclose(fp);
    return;
}


static int Strincmp( const char *s, const char *t, size_t max )
{
    for( ; max > 1; ++s, ++t, --max ) {
        if( toupper( *s ) != toupper( *t ))
            break;
        if( *s == '\0' )
            return( 0 );
    }
    return( toupper( *s ) - toupper( *t ));
}

static FILE *OpenBinfile( char *name, char *mode )
{
    FILE *fp;

    if( toupper( *mode ) == 'W' ) { /* Write access */
        if(( fp = fopen( name, OPEN_WB )) == NULL ) { 
            printf( "Can't open output file '%s'\n", name );
            exit( 1 );
        }
    } else { /* Read access */
        if(( fp = fopen( name, OPEN_RB )) == NULL ) { 
            printf( "Can't open file '%s'\n", name );
            exit( 1 );
        } 
    }
    return( fp );
}

/*___________________________________________________________________________
 |                                                                           |
 |     Main - Program                                                        |
 |                                                                           |
 |___________________________________________________________________________|
*/
int main( int argc, char *argv[] )
{
    FILE     *bit_infile, *dec_outfile;
    Word32   framecount, frameno = 0, n;

    char     InfoBitStream_r[25];
    char     tmp[1000];
    Word16   ChannelBits_r[1000];
    Word16   SynFrame[240], Bfi, cnt;


    if( sizeof( char ) != 1 || sizeof( Word16 ) != 2 || sizeof( Word32 ) != 4 ) {
        fprintf( stderr, "Wrong Datatypes : Edit Typedef.h (char, Word16, Word32)!\n" );
        return( 1 );
    }
    if( (argc < 4) || (argc > 5) ) {
        fprintf( stderr, "-----------------------------------------------------------\n");
        fprintf( stderr, "|Usage: CHAN_DEC R_ChanBitFile SynthFile ConfigFile [/f=f]|\n");
        fprintf( stderr, "-----------------------------------------------------------\n");
        fprintf( stderr, "\n");
        fprintf( stderr, "   R_ChanBitFile - channel bitstream input file\n");
        fprintf( stderr, "   SynthFile     - speech decoder output file\n");
        fprintf( stderr, "   ConfigFile    - configuration file with allowed channel bitrates\n");
        fprintf( stderr, "   /f=           - number of decoded frames (optional) \n");
        return( 1 );
    }

    bit_infile  = OpenBinfile( argv[1], "r");
    dec_outfile = OpenBinfile( argv[2], "w");


    frameno = 0;
    if(argc==5 && !Strincmp( argv[4], "/f=", 3 )){
        frameno = atoi( &argv[4][3] );
        if( frameno < 0 ) {
             printf( "frame > 0\n" );
             return( 1 );
        }
    }


    printf( "  _____________________________________________\n" );
    printf( " |                                             |\n" );
    printf( " |       G.723.1 Channel Decoder Simulation    |\n" );
    printf( " |               Version 3.1                   |\n" );
    printf( " |_____________________________________________|\n\n" );

    printf( "    Received Channel Bitstream File    : %s\n", argv[1] );
    printf( "    Synthesized Speech Signal File     : %s\n", argv[2] );
    printf( "    Channel Bitrate Configuration File : %s\n", argv[3] );
    if( frameno == 0 )
        printf( "    Simulated Frames                   : whole file\n" );
    else
        printf( "    Simulated Frames                   : %d\n", frameno );
    printf( "\n" );



    /* Read file with channel bitrates */
    /* ------------------------------- */
    GetChannelBitrates(argv[3]);


    /* Initialise G.723.1 Decoder */
    /* -------------------------- */
    Init_DecoderState ();
    Init_Decod() ;
    UsePf = True ;
    UseVx = True ;

    for( framecount = 0; framecount++ < frameno || frameno == 0; ) {

       printf( outtext, framecount);

       /* Read received soft bits */
       /* ----------------------- */
       n = fread((char*)tmp,sizeof(char),MAX_CHAN_BIT,bit_infile);
       if (n != MAX_CHAN_BIT) break;

       cnt=0; /* counter to detect not transmitted frames */
       for (n=0;n<MAX_CHAN_BIT;n++) {
          ChannelBits_r[n]=(Word16) tmp[n];
	  if (ChannelBits_r[n] == 0) cnt++;
       }

       if (cnt < MAX_CHAN_BIT) {

          /* Channel decoder */
          /* --------------- */
          ChannelDecoder( ChannelBits_r, InfoBitStream_r);

          /* G.723.1 Decoder */
          /* --------------- */
          if ((InfoBitStream_r[0] & 0x3) == 0) Bfi     = InfoBitStream_r[24] & 0x1;
          if ((InfoBitStream_r[0] & 0x3) == 1) Bfi     = InfoBitStream_r[20] & 0x1;
          if ((InfoBitStream_r[0] & 0x3) == 2) Bfi     = InfoBitStream_r[4]  & 0x1;
       }else{
          /* This frame was not transmitted and is inserted in the decoder */
          /* ------------------------------------------------------------- */
          InfoBitStream_r[0] = 0x3;
          Bfi = 0;
       }

       Decod(SynFrame, InfoBitStream_r, Bfi);

       /* Write synthesized speech signal */
       /* ------------------------------- */
       fwrite ((char*)SynFrame, sizeof(Word16), 240, dec_outfile);

    }
    printf ("\n done\n");
    fclose( bit_infile );
    fclose( dec_outfile );

    return( 0 );
}
