/*-------------------------------------------------------------------------*/ 
/*                                                                         */
/*                   G.723.1 Channel Encoder Simulation V3.1               */
/*                                                                         */
/*-------------------------------------------------------------------------*/
/*                                                                         */
/*-------------------------------------------------------------------------*/
/* (C) Copyright 1995-1996, Bosch Telecom. All rights reserved.            */
/* Contact: Joerg-Martin Mueller, Bosch Telecom, Email: jmm@bk.bosch.de    */
/*-------------------------------------------------------------------------*/


/*___________________________________________________________________________
 |                                                                           |
 | 	Include-Files                                                        |
 |___________________________________________________________________________|
*/
#include "defines.h"
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>

/* G.723.1 Includes */
#include "typedef.h"
#include "basop.h"
#include "cst_lbc.h"
#include "tab_lbc.h"
#include "lbccodec.h"
#include "coder.h"
#include "decod.h"
#include "exc_lbc.h"
#include "util_lbc.h"

#define   tstbit( x, n )    (( x ) &    ( 1 << ( n )))


/*___________________________________________________________________________
 |                                                                           |
 |  	local Functions                                                      |
 |___________________________________________________________________________|
*/
static void  GetChannelBitrates(char *ConfigName);
static void Channel_Pack(Word16 chan_bit[], char ChannelBitStream[], 
                         Word16 SysChannelBitrate, Word16 *NumChannelBytes);
static Word16 GetSpeechFrame (FILE *bit_infile, char InfoBitStream[]);
static void CrcEncoder(Word16 info_bit[], char mode);
static void ChannelEncoder( char InfoBitStream[], char ChannelBitStream[],
                            Word16 ChannelBitrate,  Word16 *NumChannelBytes);
static void bit_allocation( char InfoBitStream[], Word16 info_bit[] );
static void conv_encoder( Word16 bit_class[], Word16 punctur[5][3], 
         Word16 info_bit[], Word16 chan_bit[], char mode);
static void  EndTrellisReorderEnc (Word16 info_bit[], Word16 bit_class[], 
                                   Word16 punctur[5][3], Word16 mode);
static void ChannelCodecInit(Word16 SpeechCodecMode,Word16 ChannelBitrate,
                             Word16 bit_class[],Word16 punctur[5][3]);
static Word16 GetRate (Word16 rbits, Word16 weight, Word16 bit_class, Word16 inv_bit_class, Word16 rnd);
static bitrate_test( Word16 nbit_class[], Word16 punctur[5][3], Word16 last );
static void bitrate_adapt_high (Word16 bit_class[],Word16 punctur[5][3],
                                Word16 cbits, Word16 r[], Word16 last);
static void bitrate_adapt_low (Word16 bit_class[],Word16 punctur[5][3],
                                Word16 cbits, Word16 r[], Word16 last);
static void channel_encoder_adaptation(char InfoBitStream[],char CInfoBitStream[]);
static void Line_Pack_Channel( LINEDEF *Line, char *Vout, Word16 FrType);
static void LSP_Reorder_encoder (LINEDEF *Line);
static void SelectChannelBitrate(Word16 ChannelBitrate,Word16 *SysChannelBitrate,
                                 Word16 *CC_mode, Word16 G723_mode);
static int Strincmp( const char *s, const char *t, size_t max );
static FILE *OpenBinfile( char *name, char *mode );

/*___________________________________________________________________________
 |                                                                           |
 |  Globals                                                                  |
 |___________________________________________________________________________|
*/
/* G.723.1 declarations */
/* -------------------- */
enum  Crate    WrkRate;
Flag  UseHp = True;
Flag  UsePf = True;
Flag  UseVx = True;
Word32  count_clip;

/* Definition of CRC Check Code for error detection (see section 2.4) */
/* ------------------------------------------------------------------ */
#define   CrcLen         5  /* number of parity check bits       */
#define   CrcGen        18  /* generator polynomial              */
#define   CrcWindow53   34  /* number of bits of class 1 for crc */
#define   CrcWindow63   44  /* number of bits of class 1 for crc */
#define   CrcWindowSID  30  /* number of bits of class 1 for crc */

/* Definition of low rate configuration */
/* ------------------------------------ */
#define   NBits5_3       168  /* Info bits with configuration and UB's */
#define   NInfoBits5_3   162  /* Info bits without configuration and UB's */

/* Definition of high rate configuration */
/* ------------------------------------- */
#define   NBits6_3       200  /* Info bits with configuration and UB's */
#define   NInfoBits6_3   193  /* Info bits without configuration and UB's */

/* Definition of SID configuration */
/* ------------------------------- */
#define   NBitsSID        32  /* Info bits with configuration and UB's */
#define   NInfoBitsSID    30  /* Info bits without configuration and UB's */

/* Definition of BCH Code for configuration bits (see section 2.5.4) */
/* ----------------------------------------------------------------- */
#define   BCHLen  13
#define   NumBCHWords  8

/* Channel bitrates (are read from file) */
/* ------------------------------------- */
static Word16   ChannelBitrate_53[2];
static Word16   ChannelBitrate_63[2];
static Word16   ChannelBitrate_SID[2];

/* This flag is only true when exercising a special test sequence */
/* -------------------------------------------------------------- */
Flag TESTSEQUENCE = False;

/*___________________________________________________________________________
 |                                                                           |
 |  	Subroutines                                                          |
 |___________________________________________________________________________|
*/

/*___________________________________________________________________________
 |                                                                           |
 | Generation of channel encoder output format (see section 2.5.4)           |
 |___________________________________________________________________________|
*/
static void Channel_Pack(Word16 chan_bit[], char ChannelBitStream[], 
                         Word16 SysChannelBitrate, Word16 *NumChannelBytes)
{
   Word16 i, pw, pb;
   Word32 L_tmp;

   /* Determine number of bytes to be packed */
   /* -------------------------------------- */

   /* Multiply channel bitrate by 30./8000. */
   /* ------------------------------------- */
   L_tmp = L_mult(SysChannelBitrate,31457);
   *NumChannelBytes = (Word16) L_shr(L_tmp,24);

   /* Multiply channel bitrate by 8000./30. */
   /* ------------------------------------- */
   L_tmp = L_mult(*NumChannelBytes,17067);
   L_tmp = L_shr(L_tmp,7);
   if ((Word16) L_tmp < SysChannelBitrate) *NumChannelBytes = *NumChannelBytes + 1;

   /* Pack channel bits */
   /* ----------------- */
   for (i=0;i<(*NumChannelBytes);i++)  ChannelBitStream[i] = 0;
   pw = 0;
   pb = 0;
   for (i=0;i<*NumChannelBytes*8;i++) {
      ChannelBitStream[pw] = ChannelBitStream[pw] | (chan_bit[i] << pb);
      pb++;
      if (pb == 8) {
         pb = 0;
         pw++;
      }
   }
   return;
}

/*___________________________________________________________________________
 |                                                                           |
 | Quantization of specified channel bitrate to allowed channel bitrate      |
 |               (section 2.5.4)                                             |
 |___________________________________________________________________________|
*/
static void SelectChannelBitrate(Word16 ChannelBitrate,Word16 *SysChannelBitrate,
                                 Word16 *CC_mode, Word16 G723_mode)
{
   Word16   CC_Bitrate[2],i;
 
   /* Set allowed channel bitrates depending on G.723 mode */
   /* ---------------------------------------------------- */
   if (G723_mode == 0) {
      for (i=0;i<2;i++) CC_Bitrate[i] = ChannelBitrate_63[i];
   }
   if (G723_mode == 1) {
      for (i=0;i<2;i++) CC_Bitrate[i] = ChannelBitrate_53[i];
   }
   if (G723_mode == 2) {
      for (i=0;i<2;i++) CC_Bitrate[i] = ChannelBitrate_SID[i];
   }

   /* Quantize requested channel bitrate to next lower allowed channel bitrate */
   /* ------------------------------------------------------------------------ */
   *CC_mode = (-1);
   for (i=1;i>=0;i--) {
      if (ChannelBitrate >= CC_Bitrate[i]){
         *CC_mode = i;
         *SysChannelBitrate = CC_Bitrate[i];
         break;
      }
   }

   if (*CC_mode == (-1)){
      printf (" Requested channel bitrate is too low \n");
      exit(1);
   }      

   return;
}

/*_________________________________________________________________________________
 |                                                                                 |
 | Reorder the last protected bits to consider lower BER at the end of the trellis |
 |               (section 2.5.2)                                                   |
 |_________________________________________________________________________________|
*/
static void  EndTrellisReorderEnc (Word16 info_bit[], Word16 bit_class[], 
                                   Word16 punctur[5][3], Word16 mode)
{
   Word16 last,i,ptr_end,tmp,limit;

    if( (mode & 0x3) == 0) limit=CrcWindow63+CrcLen+20;
    if( (mode & 0x3) == 1) limit=CrcWindow53+CrcLen+20;
    if( (mode & 0x3) == 2) limit=CrcWindowSID+CrcLen+20;


   /* Check which class is the last protected one */
   /* ------------------------------------------- */
   last=0;
   for (i=1;i<5;i++){
      if (punctur[i][0] != 0) last=i;
   }

   /* Determine number of protected bits */
   /* ---------------------------------- */
   ptr_end = 0;
   for (i=0;i<=last;i++) ptr_end = ptr_end + bit_class[i];
   ptr_end = ptr_end - 4;  /* remove tail bits */

   /* Do reordering */
   /* ------------- */
   if (ptr_end >= limit) {
      ptr_end--;
      for (i=0;i<10;i++){
         tmp                    = info_bit[ptr_end-i];
         info_bit[ptr_end-i]    = info_bit[ptr_end-19+i];
         info_bit[ptr_end-19+i] = tmp;
      }
   }
   return;
}



/*_________________________________________________________________________________
 |                                                                                 |
 | Determine the parameters of the channel encoder given the channel bitrate       |
 |               (section 2.5.1)                                                   |
 |_________________________________________________________________________________|
*/
static void ChannelCodecInit(Word16 SpeechCodecMode,Word16 ChannelBitrate,
                             Word16 bit_class[],Word16 punctur[5][3])

{
/* Definition of puncturing matrix for convolutional code (see Table C4) */
/* --------------------------------------------------------------------- */
   static const Word16 PuncturTable[25][3] = {{0x000,0x000,0x000},
                                            {0xd6f,0x690,0x000},
                                            {0xd7f,0x690,0x000},
                                            {0xd7f,0x691,0x000},
                                            {0xd7f,0x695,0x000},
                                            {0xdff,0x695,0x000},
                                            {0xfff,0x695,0x000},
                                            {0xfff,0x69d,0x000},
                                            {0xfff,0x6dd,0x000},
                                            {0xfff,0x6df,0x000},
                                            {0xfff,0x7df,0x000},
                                            {0xfff,0x7ff,0x000},
                                            {0xfff,0xfff,0x000},
                                            {0xfff,0xfff,0x001},
                                            {0xfff,0xfff,0x009},
                                            {0xfff,0xfff,0x109},
                                            {0xfff,0xfff,0x309},
                                            {0xfff,0xfff,0x329},
                                            {0xfff,0xfff,0x729},
                                            {0xfff,0xfff,0x72d},
                                            {0xfff,0xfff,0x72f},
                                            {0xfff,0xfff,0x7af},
                                            {0xfff,0xfff,0x7bf},
                                            {0xfff,0xfff,0x7ff},
                                            {0xfff,0xfff,0xfff}};

/* Definition of low rate configuration (see Table C5a)         */
/* Note that the weight is in 16 bit representation and negated */
/* ------------------------------------------------------------ */
   static const Word16   BitClass53[5]    = { 34+CrcLen, 40, 40, 40+4, 8 };
   static const Word16   Weight53[5]      = { -7864, -10158, -7864, -6882, 0}; 

   static const Word16   InvBitClass53[4] = {26887, 26214, 26214, 26214}; /* 2^20/BitClass53[i]) */
   static const Word16   inv_b0p4_53      = 24386;                        /* 2^20/(BitClass53[0]+4)*/

/* Definition of high rate configuration (see Table C5a)        */
/* Note that the weight is in 16 bit representation and negated */
/* ------------------------------------------------------------ */
   static const Word16   BitClass63[5]    = { 44+CrcLen, 44, 46, 47+4, 12 };
   static const Word16   Weight63[5]      = { -8520, -9502, -7864, -6882, 0};

   static const Word16   InvBitClass63[4] = {21400, 23831, 22795, 22310};
   static const Word16   inv_b0p4_63      = 19785;

/* Definition of SID configuration (see Table C5b)              */
/* Note that the weight is in 16 bit representation and negated */
/* ------------------------------------------------------------ */
   static const Word16   BitClassSID[5] = { 30+CrcLen+4, 0, 0, 0, 0 };
   static const Word16   WeightSID[5]   = { -32768, 0, 0, 0, 0};

   static const Word16   inv_b0p4_SID   = 26887;

   Word16   i,j,rbits,r[5],abits,last,cbits,limit,NInfoBits;
   Word16   weight[5],inv_b0p4,inv_bit_class[4];
   Word32   L_tmp;

   /* Determine the number of available channel bits per frame (multiply by 0.03) */
   /* ----------------------------------------------------------------------------*/
   L_tmp = L_mult(ChannelBitrate,31458);
   cbits = (Word16) L_shr(L_tmp,21);

   if ((SpeechCodecMode & 0x3) == 0){
      for (i=0;i<5;i++) bit_class[i] = BitClass63[i];
      for (i=0;i<5;i++) weight[i] = Weight63[i];
      inv_b0p4 = inv_b0p4_63;
      for (i=0;i<4;i++) inv_bit_class[i] = InvBitClass63[i];
      NInfoBits = NInfoBits6_3;
   }
   if ((SpeechCodecMode & 0x3) == 1){
      for (i=0;i<5;i++) bit_class[i] = BitClass53[i];
      for (i=0;i<5;i++) weight[i] = Weight53[i];
      inv_b0p4 = inv_b0p4_53;
      for (i=0;i<4;i++) inv_bit_class[i] = InvBitClass53[i];
      NInfoBits = NInfoBits5_3;
   }
   if ((SpeechCodecMode & 0x3) == 2){
      for (i=0;i<5;i++) bit_class[i] = BitClassSID[i];
      for (i=0;i<5;i++) weight[i] = WeightSID[i];
      inv_b0p4 = inv_b0p4_SID;
      NInfoBits = NInfoBitsSID;
   }

   /* Determine the number of available redundancy bits per frame */
   /* ----------------------------------------------------------- */
   rbits = cbits - (NInfoBits+CrcLen+BCHLen);

   /* Check whether enough bits are available */
   /* --------------------------------------- */
   if (rbits < 0) {
      printf ("\n\nNot enough channel bitrate specified\n");
      printf ("The minimum channel bitrate is %.2f kbit/s \n",
            ((float)(BCHLen+CrcLen+NInfoBits)*1000./30.));
      exit(1);
   }

   if (rbits == 0) {
      /* No convolutional code specified */
      /* ------------------------------- */
      if (SpeechCodecMode < 2) {
         bit_class[3] = bit_class[3]-4;
      }else{
         bit_class[0] = bit_class[0]-4;
      }
      for (i=0;i<4;i++) {
         bit_class[4] = bit_class[4] + bit_class[i];
         bit_class[i] = 0;
      }
      for (i=0;i<5;i++){
         r[i]=12;
         for (j=0;j<3;j++) punctur[i][j] = PuncturTable[r[i]-12][j];
      }
   }else{

      /* Check whether too much bits have been specified */
      /* ----------------------------------------------- */
      limit = 2*(bit_class[0]+bit_class[1]+bit_class[2]+
                 bit_class[3]+bit_class[4]) + 4;
      if (rbits >= limit) {
         /* The requested channel bitrate is too high. The channel bitrate */
         /* is therefore limited to the maximum allowed value.             */
         /* -------------------------------------------------------------- */
         for (i=0;i<4;i++) r[i]=36;
         r[4]=12;
         for (i=0;i<5;i++){
            for (j=0;j<3;j++) punctur[i][j] = PuncturTable[r[i]-12][j];
         }
         bit_class[3] = bit_class[3] + bit_class[4];
         bit_class[4] = 0;
      }else{
         /* Initialise rates and puncturing tables */
         /* -------------------------------------- */
         for (i=0;i<5;i++)   r[i]=12;
         for (i=0;i<5;i++){
            for (j=0;j<3;j++) punctur[i][j] = PuncturTable[r[i]-12][j];
         }

         /* Determine rate for the most sensitive bit class */
         /* ----------------------------------------------- */
         rbits = rbits - 4; /* tail bits have to be transmitted */
         if (rbits < 0) rbits=0;
         if ((rbits <= bit_class[0]+8) || (SpeechCodecMode == 2)) {
            /* Allocate all redundancy to class 0 */
            if (SpeechCodecMode < 2) {
               bit_class[3] = bit_class[3]-4;
            }else{
               bit_class[0] = bit_class[0]-4;
            }
	    r[0] = GetRate(rbits,(Word16)(-32768),bit_class[0]+4,inv_b0p4,0);
            if (r[0] == 12) {
               for (i=0;i<4;i++) {
                  bit_class[4] = bit_class[4] + bit_class[i];
                  bit_class[i] = 0;
               }
            }else{
               bit_class[0] = bit_class[0] + 4;
            }
            for (j=0;j<3;j++) punctur[0][j] = PuncturTable[r[0]-12][j];
         }else{
	    r[0] = GetRate(rbits,weight[0],bit_class[0],inv_bit_class[0],1);
            if (r[0] < 24) r[0] = 24;
            if (r[0] > 36) r[0] = 36;
            for (j=0;j<3;j++) punctur[0][j] = PuncturTable[r[0]-12][j];
            bit_class[3] = bit_class[3] - 4;
            bit_class[0] = bit_class[0] + 4;
            for (i=1;i<4;i++) {
               abits = bitrate_test (bit_class,punctur,i-1);
               if (abits < cbits) {
		  r[i] = GetRate(rbits,weight[i],bit_class[i],inv_bit_class[i],1);
                  if (r[i] < 18)   r[i] = 18;
                  if (r[i] > 36)   r[i] = 36;
                  if (r[i] > r[0]) r[i] = r[0];
                  for (j=0;j<3;j++) punctur[i][j] = PuncturTable[r[i]-12][j];
                  bit_class[i-1] = bit_class[i-1] - 4;
                  bit_class[i]   = bit_class[i]   + 4;
               }else {
                  break;
               }
            }
         }
      }
   }


   /* Determine which class is the last protected one */
   /* ----------------------------------------------- */
   last=0;
   for (i=1;i<5;i++){
      if (r[i] > 12) last=i;
   }

   /* Determine channel bitrate with these settings */
   /* --------------------------------------------- */
   abits = bitrate_test (bit_class,punctur,last);
   if (abits > cbits) {
      /* too many bits are used: shift bits from high to low protection */
      /* -------------------------------------------------------------- */
      if (SpeechCodecMode < 2){
         bitrate_adapt_high (bit_class,punctur,cbits,r,last);
      }else{
         r[0]--;
         for (j=0;j<3;j++) punctur[0][j] = PuncturTable[r[0]-12][j];
         if (r[0] == 12) bit_class[0]=bit_class[0]-4;
      }
   }
   if ((abits < cbits) && (r[0]>12) && (rbits < limit) && (SpeechCodecMode < 2)) {
      /* too few bits are used:: shift bits from low to high protection */
      /* -------------------------------------------------------------- */
      bitrate_adapt_low (bit_class,punctur,cbits,r,last);
   }
   return;
}

/*_________________________________________________________________________________
 |                                                                                 |
 | Determine rate for one bit class                                                |
 |               (section 2.5.1)                                                   |
 |_________________________________________________________________________________|
*/
static Word16 GetRate (Word16 rbits, Word16 weight, Word16 bit_class, Word16 inv_bit_class, Word16 rnd)
{
         
   Word16 tmp;
   Word32 L_tmp1, L_tmp2;

   /* Calculate 12*(rbits*weight+bit_class) */
   /* ------------------------------------- */
   L_tmp1 = - L_mult(rbits,weight);
   L_tmp2 = - L_mult(bit_class,(Word16)(-32768));
   L_tmp1 = L_add(L_tmp1,L_tmp2);

   /* Multiply by 12 */
   /* -------------- */
   L_tmp2 = L_shl(L_tmp1,3);
   L_tmp1 = L_shl(L_tmp1,2);
   L_tmp1 = L_add(L_tmp1,L_tmp2);
   tmp   = (Word16) L_shr(L_tmp1,12);

   /* Multiply by inv_bit_class and round */
   /* ----------------------------------- */
   L_tmp1 = L_mult(tmp,inv_bit_class);
   tmp    = (Word16) L_shr(L_tmp1,24);
   tmp    = add(tmp,rnd);
   tmp    = shr (tmp,1);
   return(tmp);
}

/*_________________________________________________________________________________
 |                                                                                 |
 | Adaptation of bit allocation to meet channel bitrate requirement                |
 |               (section 2.5.1)                                                   |
 |_________________________________________________________________________________|
*/
static void bitrate_adapt_low (Word16 bit_class[],Word16 punctur[5][3],
                               Word16 cbits, Word16 r[], Word16 last)
{
   Word16 abits,src,dest;

   src  = last+1;
   dest = last;
   for (;;) {
      if (bit_class[src] == 0) {
         if (src == last) return;
         src = last;
         dest = last-1;
         if (r[src] == r[dest]) return;
      }
      /* shift one bit from bit_class[src] to bit_class[dest] */
      /* ---------------------------------------------------- */
      bit_class[src]--;
      bit_class[dest]++;
      abits = bitrate_test (bit_class,punctur,last);
      if (abits >= cbits) break;      
   }

   if (abits > cbits) bitrate_adapt_high (bit_class,punctur,cbits,r,last);

   return;
}

/*_________________________________________________________________________________
 |                                                                                 |
 | Adaptation of bit allocation to meet channel bitrate requirement                |
 |               (section 2.5.1)                                                   |
 |_________________________________________________________________________________|
*/
static void bitrate_adapt_high (Word16 bit_class[],Word16 punctur[5][3],
                                Word16 cbits, Word16 r[], Word16 last)
{
   Word16 abits,i;

   if (r[3] > 21) {
      /* The number of bits in the unprotected class may not be increased */
      /* ---------------------------------------------------------------- */
      for (;;) {
         for (i=2;i>0;i--){
            /* shift one bit of bit_class[i] to bit_class[i+1] */
            /* ----------------------------------------------- */
            if (bit_class[i] == 0) return;
            bit_class[i]--;
            bit_class[i+1]++;
            abits = bitrate_test (bit_class,punctur,last);
            if (abits <= cbits) break;      
         }
         if (abits <= cbits) break;      
      }
   }else{
      for (;;) {
         /* shift one bit of bit_class[last] to bit_class[last+1] */
         /* ----------------------------------------------------- */
         if (bit_class[last] == 0) {
            printf ("\n\n Error in bitrate_adapt_high\n\n");
            exit(1);
         }
         bit_class[last]--;
         bit_class[last+1]++;
         abits = bitrate_test (bit_class,punctur,last);
         if (abits <= cbits) break;
      }
   }
   return;
}


/*_______________________________________________________________________
 |                                                                      |
 | Determination of required channel bitrate                            |
 |               (section 2.5.1)                                        |
 |______________________________________________________________________|
*/
static bitrate_test( Word16 nbit_class[], Word16 punctur[5][3], Word16 last )
{
    Word16 i, j, k, ptr0, cnt;
    static const Word16 mask[16] = { 0X0001, 0X0002, 0X0004, 0X0008,
                                     0X0010, 0X0020, 0X0040, 0X0080,
                                     0X0100, 0X0200, 0X0400, 0X0800,
                                     0X1000, 0X2000, 0X4000, -32768 };

    /* Encode the protected classes: */
    /* ----------------------------- */
    ptr0 = 0;
    cnt  = 0;
    for( k = 0; k <= last; ++k ) {
        for( i = 0; i < nbit_class[k]; ++i ) {
            for( j = 0; j < 3; ++j ) {
                if(( punctur[k][j] & mask[cnt] ) != 0 ) ptr0++;
            }
            cnt = ++cnt % 12;
        }
    }

    /* Determine channel bitrate */
    /* ------------------------- */
    for (i=last+1;i<5;i++) ptr0 = ptr0 + nbit_class[i];

    ptr0 += BCHLen;

    return(ptr0);

}

/*__________________________________________________________________________
 |                                                                          |
 | Reorder G.723.1 information bitstream considering subjective sensitivity |
 |               (sections 2.2 and 2.3)                                     |
 |__________________________________________________________________________|
*/
static void bit_allocation( char CInfoBitStream[], Word16 info_bit[] )
{

/* Definition of high rate bitordering table (see Table C3a) */
/* --------------------------------------------------------- */
static const Word16   BitOrder63[NBits6_3] = 
       {175, 180, 189, 190, 191, 192 , -1 , -1,
         98 , 73, 107, 154, 167, 168, 169, 170,
         30 , 17 , 16 , 31 , 48 , 55 , 49 , 71,
          6  , 4  , 0  , 2 , 11 , 26 , 10 , 14,
          5  , 1  , 3 , 12 , 27 , 24 , 60  , 8,
         44 , 66 , 62 , 82 , 25 , 61  , 9  , 7,
         45 , 67 , 63 , 83 , 78 , 36 , 50 , 40,
         46 , 68 , 64 , 84 , 79 , 37 , 51 , 41,
         47 , 69 , 65 , 85 , 80 , 38 , 52 , 42,
         56 , 99, 159, 185 , 81 , 39 , 53 , 43,
        161, 187 , 20,  57 ,100, 160, 186 , 19,
         22,  59 ,102, 162, 188 , 21,  58 ,101,
         35 , 54 , 70 , 72, 179, 178, 177, 176,
         13 , 15 , 23 , 28 , 29 , 32 , 33 , 34,
         128, 132, 146, 155, 163, 171, 181, 18,
         76 , 86 , 90 , 94 ,103, 108, 112, 116,
        129, 133, 147, 156, 164, 172, 182 , 74,
        183 , 87 , 91 , 95, 104, 109, 113, 117,
        114, 118, 130, 134, 148, 157, 165, 173,
        184 , 75 , 77 , 88 , 92 , 96, 105, 110,
        115, 119, 131, 135, 149, 158, 166, 174,
        136, 124, 120 , 89 , 93 , 97, 106, 111,
        151, 141, 137, 125, 121, 144, 150, 140,
        127, 123, 145, 152, 142, 138, 126, 122,
         -1 , -1 , -1 , -1 , -1, 153, 143, 139};

/* Definition of low rate bitordering table (see Table C3b) */
/* -------------------------------------------------------- */
static const Word16   BitOrder53[NBits5_3] = 
      {152, 153, 158, 159, 160, 161 , -1 , -1,
        69 , 64 , 70 , 91, 145, 140, 147, 146 ,
        24 , 15 , 14 , 25 , 46 , 50 , 47,  63,
         4  , 6  , 0  , 2 , 11 , 18 , 10,  13 ,
         7  , 1  , 3 , 12 , 19 , 16 , 48,   8,
        42 , 59 , 55 , 65 , 17 , 49  , 9,   5   ,
        43 , 60 , 56 , 66 , 26 , 30 , 34,  38  ,
        44 , 61 , 57 , 67 , 27 , 31 , 35,  39  ,
        45 , 62 , 58 , 68 , 28 , 32 , 36,  40  ,
        51 , 87 ,141, 154 , 29 , 33 , 37,  41  ,
       143, 156 , 21,  52 , 88 ,142, 155,  20 ,
        23,  54 , 90 ,144, 157 , 22,  53,  89  ,
       100, 128 , 96, 104, 151, 150, 149, 148,
       132, 112, 116, 136, 108, 120, 124,  92 ,
       109, 121, 125 , 93, 101, 129 , 97, 105,
       102, 130 , 98, 106, 133, 113, 117, 137 ,
       134, 114, 118, 138, 110, 122, 126,  94 ,
       111, 123, 127 , 95, 103, 131 , 99, 107,
        83 , 79 , 75 , 71, 135, 115, 119, 139 ,
        85 , 81 , 77 , 73 , 84 , 80 , 76,  72,
        -1 , -1 , -1 , -1 , 86 , 82 , 78,  74};

/* Definition of SID frame bitordering table (see Table C3c) */
/* --------------------------------------------------------- */
static const Word16   BitOrderSID[NBitsSID] = 
       {  21, 22, 23, 24, 25, 26, -1, -1,
          13, 14, 15, 16, 17, 18, 20, 19,
           5,  6,  7,  8,  9, 10, 11, 12,
           0,  1,  2, 29, 28, 27,  3,  4};


    Word16    i, pw, pb, j;
    Word16    BitOrder[NBits6_3], EndLoop;

    if( (CInfoBitStream[0] & 0x3) == 0) {   
       EndLoop = NBits6_3;
       for (i=0;i<EndLoop;i++) BitOrder[i] = BitOrder63[i];
    }
    if( (CInfoBitStream[0] & 0x3) == 1) {   
       EndLoop = NBits5_3;
       for (i=0;i<EndLoop;i++) BitOrder[i] = BitOrder53[i];
    } 
    if( (CInfoBitStream[0] & 0x3) == 2) {   
       EndLoop = NBitsSID;
       for (i=0;i<EndLoop;i++) BitOrder[i] = BitOrderSID[i];
    }


    /* Reorder */
    /* ------- */
    for( i=0; i < EndLoop; i++ ) {
       pw = i/8;
       pb = i - pw*8;
       j  = (pw*8)+7-pb;
       if (BitOrder[j] != (-1)){
          info_bit[BitOrder[j]] = ( tstbit( CInfoBitStream[pw], pb )) ? 1 : 0;
       }
    }       
    return;
}

/*__________________________________________________________________________
 |                                                                          |
 |      Determine CRC bits for most sensitive bitclass 0                    |
 |               (section 2.4)                                              |
 |__________________________________________________________________________|
*/
static void   CrcEncoder(Word16 info_bit[], char mode)
{

   Word16        crc_window, EndLoop;
   Word16        CrcCarry, ShiftReg, i;

    if( (mode & 0x3) == 0) {
      crc_window = CrcWindow63;
      EndLoop = NInfoBits6_3;
    }
    if( (mode & 0x3) == 1) {   
      crc_window = CrcWindow53;
      EndLoop = NInfoBits5_3;
    } 
    if( (mode & 0x3) == 2) {
      crc_window = CrcWindowSID;
      EndLoop = NInfoBitsSID;
    }

   /* CRC encoder */
   /* ----------- */
   ShiftReg = 0;
   for (i = 0; i < crc_window; ++i) {
     CrcCarry = (info_bit[i] ^ ShiftReg) & 1;
     ShiftReg = shr(ShiftReg, 1);
     if (CrcCarry != 0) ShiftReg = ShiftReg ^ CrcGen;
   }

   /* Make room for CRC bits */
   /* ---------------------- */
   for (i=EndLoop-1;i>=crc_window;i--) info_bit[i+CrcLen]=info_bit[i];

   /* Copy CRC bits */
   /* ------------- */
   for (i=CrcLen-1;i>=0;i--){
      info_bit[crc_window + i] = ShiftReg & 1;
      ShiftReg = shr(ShiftReg, 1);
   }

   return;
}

/*__________________________________________________________________________
 |                                                                          |
 | Enhancement of G.723.1 bitstream robustness                              |
 |               (section 2.1)                                              |
 |__________________________________________________________________________|
*/
static void channel_encoder_adaptation(char InfoBitStream[],char CInfoBitStream[])
{
   Word16   FrType, i, Ftyp;
   LINEDEF  Line ;

   FrType = InfoBitStream[0] & 3;
   if (FrType < 2) {
      /* Unpack the G.723.1 bitstream */
      /* ---------------------------- */
      Line = Line_Unpk( InfoBitStream, &Ftyp, (Word16) 0 ) ;

      /* Reorder LSP index for enhanced robustness */
      /* ----------------------------------------- */
      LSP_Reorder_encoder (&Line);

      /* Demultiplex gain indexes and repack the robust bitstream */
      /* -------------------------------------------------------- */
      Line_Pack_Channel( &Line, CInfoBitStream, FrType ) ;
   }else{
      /* No reordering for SID frames */
      /* ---------------------------- */
      for (i=0;i<4;i++) CInfoBitStream[i] = InfoBitStream[i];
   }
   return;
}

/*__________________________________________________________________________
 |                                                                          |
 |          Reordering of LSP index                                         |
 |               (section 2.1.2)                                            |
 |__________________________________________________________________________|
*/
static void LSP_Reorder_encoder (LINEDEF *Line)
{

/* Reordering table for first subvector (see Table C1a) */
/* ---------------------------------------------------- */
static	Word16 tabd2_01[LspCbSize] = {
	  82,   91,  190,  191,  189,   36,  187,   32, 
	  38,   39,  185,  112,  166,  175,  116,  120, 
	  34,   35,  122,   41,   40,   43,   42,   56, 
	  57,   90,  168,   85,   74,  170,  234,  174, 
	 169,  172,  178,  182,  184,  179,  181,  180, 
	  37,  186,   44,   33,  159,  183,  188,  155, 
	 253,  252,  147,  154,  246,  165,  218,  139, 
	 163,  160,  157,  146,  145,  144,  177,  176, 
	 244,  131,  148,  129,  128,  161,  219,  135, 
	 134,  203,  200,  206,  207,  204,  205,  254, 
	 212,  222,  213,  220,  221,  141,  216,   88, 
	 138,  137,  136,  217,  133,  132,  201,  197, 
	 196,   76,   77,  243,  192,  195,  193,  119, 
	 118,  108,  121,  249,  247,  245,  130,  240, 
	 241,  235,   69,   68,  210,  226,  227,  224, 
	 225,   79,  211,  250,  251,  255,  198,  242, 
	 194,  127,   45,   46,  117,  125,  124,    2, 
	  67,    3,  101,   72,    0,   13,   97,   65, 
	  66,   98,   96,  158,  248,  228,  229,   93, 
	 231,  109,  100,  110,  111,   75,   73,  115, 
	 114,   23,   94,   95,    6,   22,    7,  230, 
	   5,    4,   99,   81,   20,   21,   83,  113, 
	 102,   14,  103,   15,   12,   16,   52,   17, 
	  19,    1,   29,    8,   28,   30,   31,   11, 
	   9,  126,   24,   18,   26,   25,   27,   86, 
	  87,   64,   71,   70,  153,  152,  173,  162, 
	 167,  164,   89,   80,  238,  208,  232,  233, 
	 215,  236,  239,   92,  202,  209,  199,  143, 
	 142,  140,  223,   84,   78,  171,  151,  156, 
	 237,  150,  149,  214,  106,  107,   10,   53, 
	  47,   54,   55,  104,  105,  123,   48,   58, 
	  59,   49,   50,   51,   62,   60,   61,   63
	};


/* Reordering table for second subvector (see Table C1b) */
/* ----------------------------------------------------- */
static	Word16 tabd1_01[LspCbSize] = {
	 122,  150,  201,  213,  214,  212,  246,  244, 
	 136,  135,  130,  134,  140,  216,  218,  146, 
	 209,  208,  110,  153,  156,  158,  152,  144, 
	 147,  145,  199,  197,  193,  204,  205,  249, 
	 248,  127,  185,  172,  174,  173,  179,  169, 
	 168,  181,  182,  177,  183,  161,  178,  167, 
	 163,  162,  160,  165,  232,  234,  235,  233, 
	 237,  236,  238,  239,   78,  155,  154,   75, 
	  74,   69,   68,   71,   70,   73,   72,  139, 
	 123,  114,   59,   58,  113,  170,  137,  141, 
	 186,  133,  151,  132,  148,  159,  196,  149, 
	 187,  189,  191,  157,  198,  190,  121,  126, 
	 125,  120,   57,  112,  171,   63,   45,   62, 
	  47,   46,   99,  109,   98,  103,  124,  102, 
	 107,  106,  129,  128,  131,   89,  118,  105, 
	 104,   95,   91,   90,   94,   86,   83,   82, 
	 138,  119,  117,   88,  116,   92,   37,   36, 
	  38,   32,   96,   87,   81,   39,   93,   49, 
	  48,   23,   22,   21,   20,   29,    8,   56, 
	  55,   44,  175,   60,   61,   53,   43,  143, 
	 115,  176,  180,  142,  166,  184,  188,  207, 
	 206,  231,  252,  195,  194,  219,  217,  221, 
	 223,  211,  215,  210,  192,  200,  164,  226, 
	 202,  203,  230,  227,  229,  253,  255,  224, 
	 225,  243,  228,  220,  240,  241,  242,  222, 
	 245,  251,  250,  247,  254,  111,  108,   77, 
	  76,   79,  101,  100,   30,   15,   97,   52, 
	  54,   41,   50,   51,   40,   33,   42,   35, 
	  34,   13,   12,   14,   85,   84,   25,   31, 
	  27,   80,   17,   67,   66,   16,    9,   28, 
	  24,   65,   11,   10,   26,   64,   19,   18, 
	   4,    1,    0,    5,    7,    6,    3,    2
	};


/* Reordering table for third subvector (see Table C1c) */
/* ---------------------------------------------------- */
static	Word16 tabd0_01[LspCbSize] = {
	 243,  154,  190,  170,  189,  169,  140,  188, 
	 171,  168,  166,  164,  165,  216,  146,  132, 
	 242,  234,  158,  144,  252,  147,  148,   13, 
	  74,   75,   19,   77,   76,   79,   78,   70, 
	  71,  198,  199,   72,   12,   73,  220,  150, 
	 151,  131,  204,  133,  205,  230,  229,  225, 
	 200,  201,  226,  227,  196,  192,  197,  195, 
	 202,  203,  136,  143,  142,  175,  207,  206, 
	 137,  138,  139,  174,  173,  172,  240,  228, 
	 209,  224,  238,  239,  128,  130,  145,  250, 
	 236,  237,  255,  248,  253,  232,  153,  179, 
	 181,  180,  235,  233,  152,  156,  184,  178, 
	 185,  162,  183,  182,  163,  160,  161,  167, 
	 157,  159,  155,  134,  186,  191,  135,  187, 
	 141,  254,  177,  251,  176,   33,  244,  245, 
	  61,   60,   34,   35,   32,  249,  126,   41, 
	  62,   59,   58,   63,   57,   56,   42,   43, 
	  25,   29,   40,  125,  120,   24,   26,  127, 
	 123,  124,  122,   48,  121,   46,   47,  105, 
	 104,  109,  108,   28,  116,  112,  113,  247, 
	 246,   36,   54,  149,   55,    4,   52,   53, 
	  49,   37,   39,   51,   50,  107,  211,  215, 
	  20,   21,   17,    3,    5,  223,  222,    0, 
	 231,  129,  218,    9,  217,  219,  194,  221, 
	 193,  210,   38,    2,    1,  212,   14,   15, 
	 241,   44,  208,    8,   45,   11,   10,  110, 
	  68,   69,  103,  100,  213,  214,   16,  106, 
	  18,  111,   22,   23,    6,    7,   64,   65, 
	  27,  118,  117,  102,   31,   30,   97,  114, 
	 115,   96,   88,   89,   90,   91,   99,   94, 
	  80,   82,   83,  119,   98,  101,   92,   95, 
	  66,   67,   84,   93,   85,   81,   86,   87
	};

Word16   lsp0,lsp1,lsp2;

/* Extract and reorder Band 2 */
/* -------------------------- */
   lsp2 = (Word16) (Line->LspId & (Word32) 0x000000FF);
   lsp2 = tabd2_01[lsp2];

/* Extract and reorder Band 1 */
/* -------------------------- */
   lsp1 = (Word16) ((Line->LspId >> 8) & (Word32) 0x000000FF);
   lsp1 = tabd1_01[lsp1];

/* Extract and reorder Band 0 */
/* -------------------------- */
   lsp0 = (Word16) ((Line->LspId >> 16) & (Word32) 0x000000FF);
   lsp0 = tabd0_01[lsp0];

/* Construct new Line.LspId */
/* ------------------------ */
   Line->LspId = (Word32) lsp0;
   Line->LspId = (Line->LspId << 8) + (Word32) lsp1;
   Line->LspId = (Line->LspId << 8) + (Word32) lsp2;

   return;
}


/*__________________________________________________________________________
 |                                                                          |
 | Demultiplexing of gain indexes and repacking of new information bitstream|
 |               (section 2.1.3)                                            |
 |__________________________________________________________________________|
*/
static void  Line_Pack_Channel( LINEDEF *Line, char *Vout, Word16 FrType)
{
	int		i ;
	int		BitCount ;

	Word16	InfoBitStream_tmp[195] ;
	Word16    *Bsp = InfoBitStream_tmp ;
	Word32	Temp ;
        static Word16  MapIndexMamp[24]={0,1,2,3,4,5,6, 7, 8, 9,10,11,12,13,
                                       14,15,16,17,18,19,20,21,22,23};

	/* Clear the output vector */
	for ( i = 0 ; i < 25 ; i ++ )
		Vout[i] = 0 ;

	/* 
	Add the coder rate info and the VAD status to the 2 msb
        of the first word of the frame.

	The signalling is as follows:
        00  :   High Rate
        01  :   Low  Rate
        10  :   Silence frame
        11  :   Reserved for future use
	*/

	Temp = 2L ;
	if ( FrType < 2 ) {
		if ( FrType == 0)
			Temp = 0x00000000L ;
		else
			Temp = 0x00000001L ;
	}

	/* Serialize Control info */
	Bsp = Par2Ser( Temp, Bsp, 2 ) ;

	/* 24 bit LspId */
	Temp = (*Line).LspId ;
	Bsp = Par2Ser( Temp, Bsp, 24 ) ;

	/* Check for Speech/NonSpeech case */
	if ( FrType < 2 ) {

		/*
		 	Do the part common to both rates
		*/

		/* Adaptive code book lags */
		Temp = (Word32) (*Line).Olp[0] - (Word32) PitchMin ;
		Bsp = Par2Ser( Temp, Bsp, 7 ) ;

		Temp = (Word32) (*Line).Sfs[1].AcLg ;
		Bsp = Par2Ser( Temp, Bsp, 2 ) ;

		Temp = (Word32) (*Line).Olp[1] - (Word32) PitchMin ;
		Bsp = Par2Ser( Temp, Bsp, 7 ) ;

		Temp = (Word32) (*Line).Sfs[3].AcLg ;
		Bsp = Par2Ser( Temp, Bsp, 2 ) ;

		/* Write 8 bit index of adaptive gains */
                /* ----------------------------------- */
		for ( i = 0 ; i < SubFrames ; i ++ ) {
                  Temp = (*Line).Sfs[i].AcGn;
                  if ((FrType == 0 )&&((*Line).Olp[i>>1] < (SubFrLen-2) )) {
                     Temp += (Word32) (*Line).Sfs[i].Tran << 7 ;
                  }
                  Bsp = Par2Ser( Temp, Bsp, 8 ) ;
		}

		/* Write 5 bit index of fixed gains */
                /* -------------------------------- */
		for ( i = 0 ; i < SubFrames ; i ++ ) {
                  Temp = (*Line).Sfs[i].Mamp ;
                  Temp = MapIndexMamp[Temp]; /* map the index to take maximum
                                                advantage of error detection */
                  Bsp = Par2Ser( Temp, Bsp, 5 ) ;
		}

		/* Write all the Grid indices */
		for ( i = 0 ; i < SubFrames ; i ++ )
			*Bsp ++ = (*Line).Sfs[i].Grid ;

		/* High rate only part */
		if ( FrType == 0 ) {

			/* Write 13 bit combined position index */
			Temp = (*Line).Sfs[0].Ppos >> 16 ;
			Temp = Temp * 9 + ( (*Line).Sfs[1].Ppos >> 14) ;
			Temp *= 90 ;
			Temp += ((*Line).Sfs[2].Ppos >> 16) * 9 + ( (*Line).Sfs[3].Ppos >> 14 ) ;
			Bsp = Par2Ser( Temp, Bsp, 13 ) ;

			/* Write all the pulse positions */
			Temp = (*Line).Sfs[0].Ppos & 0x0000ffffL ;
			Bsp = Par2Ser( Temp, Bsp, 16 ) ;

			Temp = (*Line).Sfs[1].Ppos & 0x00003fffL ;
			Bsp = Par2Ser( Temp, Bsp, 14 ) ;

			Temp = (*Line).Sfs[2].Ppos & 0x0000ffffL ;
			Bsp = Par2Ser( Temp, Bsp, 16 ) ;

			Temp = (*Line).Sfs[3].Ppos & 0x00003fffL ;
			Bsp = Par2Ser( Temp, Bsp, 14 ) ;

			/* Write pulse amplitudes */
			Temp = (Word32) (*Line).Sfs[0].Pamp ;
			Bsp = Par2Ser( Temp, Bsp, 6 ) ;

			Temp = (Word32) (*Line).Sfs[1].Pamp ;
			Bsp = Par2Ser( Temp, Bsp, 5 ) ;

			Temp = (Word32) (*Line).Sfs[2].Pamp ;
			Bsp = Par2Ser( Temp, Bsp, 6 ) ;

			Temp = (Word32) (*Line).Sfs[3].Pamp ;
			Bsp = Par2Ser( Temp, Bsp, 5 ) ;
		}
		/* Low rate only part */
		else {

			/* Write 12 bits of positions */
			for ( i = 0 ; i < SubFrames ; i ++ ) {
				Temp = (*Line).Sfs[i].Ppos ;
				Bsp = Par2Ser( Temp, Bsp, 12 ) ;
			}

			/* Write 4 bit Pamps */
			for ( i = 0 ; i < SubFrames ; i ++ ) {
				Temp = (*Line).Sfs[i].Pamp ;
				Bsp = Par2Ser( Temp, Bsp, 4 ) ;
			}
		}

	}
	else {
		/* Do Sid frame gain */
		
	}

	/* Write out voiced frames */
	if ( FrType == 0 )   BitCount = 192+3 ;
	if ( FrType == 1 )   BitCount = 160+4 ;
	if ( FrType == 2 )   BitCount = 32 ;

	for ( i = 0 ; i < BitCount ; i ++ )
		Vout[i>>3] ^= InfoBitStream_tmp[i] << (i & 0x0007) ;
}

/*__________________________________________________________________________
 |                                                                          |
 |          Channel encoder functions                                       |       
 |               (section 2)                                                |
 |__________________________________________________________________________|
*/
static void ChannelEncoder( char InfoBitStream[], char ChannelBitStream[],
                            Word16 ChannelBitrate,  Word16 *NumChannelBytes)
{

/* Definition of BCH Code for configuration bits (see section 2.5.4) */
/* ----------------------------------------------------------------- */
static const Word16 ConfigGenMatrix[NumBCHWords][BCHLen] = {
{ 127, 127, 127, 127, 127, 127, 127, 127, 127, 127, 127, 127, 127},
{-127,-127,-127, 127,-127,-127, 127, 127,-127, 127,-127, 127, 127},
{ 127,-127,-127,-127, 127,-127,-127, 127, 127,-127, 127,-127, 127},
{-127, 127, 127,-127,-127, 127,-127, 127,-127,-127,-127,-127, 127},
{ 127, 127,-127,-127,-127, 127,-127,-127, 127, 127,-127, 127,-127},
{-127,-127, 127,-127, 127,-127,-127,-127,-127, 127, 127, 127,-127},
{ 127,-127, 127, 127,-127,-127, 127,-127, 127,-127,-127,-127,-127},
{-127, 127,-127, 127, 127, 127, 127,-127,-127,-127, 127,-127,-127}};

    static Word16 LastMode = {-1};
    static Word16 bit_class[5],punctur[5][3];
    Word16        info_bit[210], i, mode, G723_mode, CC_mode;
    Word16        chan_bit[700], SysChannelBitrate;
    char          CInfoBitStream[25];


    G723_mode = InfoBitStream[0] & 3;
    if (TESTSEQUENCE == True) {
       /* This is only for processing the test sequences and not an allowed  */
       /* operational mode:                                                  */
       /* The requested channel bitrate is not mapped to the allowed channel */
       /* bitrates. This enables to exercise all possible channel bitrates   */
       /* and to verify the channel bitrate allocation algorithm             */
       /* ------------------------------------------------------------------ */
       SysChannelBitrate = ChannelBitrate;
       CC_mode = 0;
       /* Limit channel bitrate to maximum */
       /* -------------------------------- */
       if ((G723_mode == 0) && (SysChannelBitrate > 20634)) SysChannelBitrate = 20634;
       if ((G723_mode == 1) && (SysChannelBitrate > 17534)) SysChannelBitrate = 17534;
       if ((G723_mode == 2) && (SysChannelBitrate >  4334)) SysChannelBitrate =  4334;
    }else{
       /*          This is the normal operational mode:                  */
       /* Get allowed channel bitrate from channel bitrate table (section 2.5.4) */
       /* ---------------------------------------------------------------------- */
       SelectChannelBitrate(ChannelBitrate,&SysChannelBitrate,&CC_mode,G723_mode);
    }
    mode = (CC_mode << 2) + G723_mode;


    if (mode != LastMode){
       /* (Re-)Initialise Channel Codec (section 2.5.1) */
       /* --------------------------------------------- */
       ChannelCodecInit(G723_mode,SysChannelBitrate,bit_class,punctur);
       LastMode = mode;
    }

    /* Increase robustness by separating gain indexes and reordering LSP's (section 2.1) */
    /* --------------------------------------------------------------------------------- */
    channel_encoder_adaptation(InfoBitStream,CInfoBitStream);


    /* Allocate information bits to protection classes (sections 2.2 and 2.3) */
    /* ---------------------------------------------------------------------- */
    bit_allocation(CInfoBitStream, &info_bit[4] );


    /* Calculate CRC for class 0 and add to infobit-stream (section 2.4) */
    /* ----------------------------------------------------------------- */
    CrcEncoder(&info_bit[4],CInfoBitStream[0]);


    /* Consider low residual error probability at the end of the trellis (section 2.5.2) */
    /* --------------------------------------------------------------------------------- */
    EndTrellisReorderEnc (&info_bit[4], bit_class, punctur, CInfoBitStream[0]);


    /* Convolutionally encode (section 2.5.3) */
    /* -------------------------------------- */
    conv_encoder(bit_class, punctur, &info_bit[4], chan_bit, CInfoBitStream[0]);


    /* Encode configuration bits (section 2.5.4) */
    /* ----------------------------------------- */
    for (i=0;i<BCHLen;i++) 
       chan_bit[i] = (ConfigGenMatrix[mode][i] & 0x0080) >> 7;


    /* Pack channel bits (section 2.5.4) */
    /* --------------------------------- */
    Channel_Pack(chan_bit,ChannelBitStream,SysChannelBitrate,NumChannelBytes);

    return;
}

/*__________________________________________________________________________
 |                                                                          |
 |           Convolutional encoder                                          |       
 |               (section 2.5.3)                                            |
 |__________________________________________________________________________|
*/
static void conv_encoder( Word16 nbit_class[], Word16 punctur[5][3], 
                          Word16 info_bit[], Word16 chan_bit[], char mode)
{
    Word16 i, j, k, ptr0, ptr1, temp, ninfo_bits, cnt, out[3], last, n0;
    Word16 tmp[193];
    static const Word16 mask[16] = { 0X0001, 0X0002, 0X0004, 0X0008,
                                   0X0010, 0X0020, 0X0040, 0X0080,
                                   0X0100, 0X0200, 0X0400, 0X0800,
                                   0X1000, 0X2000, 0X4000, -32768 };


    if( (mode & 0x3) == 0) {   
       ninfo_bits = NInfoBits6_3+CrcLen;
       if ((nbit_class[0]+nbit_class[1]+
           nbit_class[2]+nbit_class[3]+nbit_class[4]) > (NInfoBits6_3+CrcLen+4)){
          printf ("\n\n\nNumber of info bits incorrect! \n");
          exit(1);
       }
    }
    if( (mode & 0x3) == 1) {   
       ninfo_bits = NInfoBits5_3+CrcLen;
       if ((nbit_class[0]+nbit_class[1]+
           nbit_class[2]+nbit_class[3]+nbit_class[4]) > (NInfoBits5_3+CrcLen+4)){
          printf ("\n\n\nNumber of info bits incorrect! \n");
          exit(1);
       }
    } 
    if( (mode & 0x3) == 2) {   
       ninfo_bits = NInfoBitsSID+CrcLen;
       if ((nbit_class[0]+nbit_class[1]+
           nbit_class[2]+nbit_class[3]+nbit_class[4]) > (NInfoBitsSID+CrcLen+4)){
          printf ("\n\n\nNumber of info bits incorrect! \n");
          exit(1);
       }
    }

    if (punctur[0][0] != 0) {

       for( i = -4; i < 0; ++i ) info_bit[i] = 0;

       /* Check which class is the last protected one */
       /* ------------------------------------------- */
       last = 0;
       for (i=1;i<5;i++){
          if (punctur[i][0] != 0) last=i;
       }


       /* save not protected bits: */
       /* ------------------------ */
       for (n0=0,i=last+1;i<5;i++) n0 += nbit_class[i];
       ptr0 = 0;
       for( i = ninfo_bits - n0; i < ninfo_bits; ++i ) {
            tmp[ptr0++] = info_bit[i];
       }

       /* Clear tailbits: */
       /* --------------- */
       for( i = ninfo_bits - n0; i < ninfo_bits - n0 + 4; ++i ) {
           info_bit[i] = 0;
       }

       /* Encode protected classes: */
       /* ------------------------- */
       ptr1 = 0;
       ptr0 = BCHLen;
       cnt  = 0;
       for( k = 0; k <= last; ++k ) {
           for( i = 0; i < nbit_class[k]; ++i ) {
               /* Determine the output of the encoder. The generator */
               /* polynomial is (23,35,27) octal (section 2.5)       */
               /* -------------------------------------------------- */
               temp   = info_bit[ptr1] + info_bit[ptr1-4];
               out[0] = temp + info_bit[ptr1-1];
               out[1] = temp + info_bit[ptr1-2] + info_bit[ptr1-3];
               out[2] = temp + info_bit[ptr1-1] + info_bit[ptr1-2];
               for( j = 0; j < 3; ++j ) {
                   if(( punctur[k][j] & mask[cnt] ) != 0 ) {
                       chan_bit[ptr0++] = out[j] & 1;
                   }
               }
               ++ptr1;
               cnt = ++cnt % 12;
           }
       }

       /* Copy not protected bits: */
       /* ------------------------ */
       for( i = 0; i < n0; ++i ) {
            chan_bit[ptr0++] = tmp[i];
       }


    }else {

       /* no convolutional code has been specified */
       /* ---------------------------------------- */
       ptr0 = BCHLen;
       for( i = 0; i < nbit_class[4]; ++i ) {
            chan_bit[ptr0++] = info_bit[i];
       }
    }

    /* Set eventually unused bits to zero */
    /* ---------------------------------- */
    for (i=ptr0;i<700;i++) chan_bit[i]=0;

    return;
}

/*__________________________________________________________________________
 |                                                                          |
 |        Read one G.723.1 frame from file                                  |       
 |                                                                          |
 |__________________________________________________________________________|
*/
static Word16 GetSpeechFrame (FILE *bit_infile, char InfoBitStream[])
{
   Word16 nread, n, FrType;

   /* Read one byte from bitstream file to find out G.723.1 mode */
   /* ---------------------------------------------------------- */
   n = fread((char*)InfoBitStream,sizeof(char),1,bit_infile);
   if (n != 1) return(1);

   FrType = (Word16) InfoBitStream[0] & 0x3;

   if (FrType == 0) nread = 23;
   if (FrType == 1) nread = 19;
   if (FrType == 2) nread =  3;
   if (FrType == 3) return(0);

   /* Read the remaining bytes from bitstream file */
   /* -------------------------------------------- */
   n = fread((char*)&InfoBitStream[1],sizeof(char),nread,bit_infile);
   if (n != nread) return(1);

  return(0);
}

/*__________________________________________________________________________
 |                                                                          |
 |        Read allowed channel bitrates from file                           |       
 |                                                                          |
 |__________________________________________________________________________|
*/
static void  GetChannelBitrates(char *ConfigName)
{
    FILE   *fp;
    Word32   flag, i, tmp;
    char   InputLine[100];

    if((fp = fopen(ConfigName,"r"))==NULL){
       fprintf(stderr,"Can't open file '%s'\n",ConfigName);
       exit(1);
    }
    flag=0;
    while(flag == 0 ) {
      fgets( InputLine, 100, fp);
      if( InputLine[0] != '#') flag=1;
    }
    sscanf(InputLine, "%d", &tmp);
    ChannelBitrate_53[0] = (Word16) tmp;
    for (i=1;i<2;i++) {
      fgets( InputLine, 100, fp);
      sscanf(InputLine, "%d", &tmp);
      ChannelBitrate_53[i] = (Word16) tmp;
    }

    flag=0;
    while(flag == 0 ) {
      fgets( InputLine, 100, fp);
      if( InputLine[0] != '#') flag=1;
    }
    sscanf(InputLine, "%d", &tmp);
    ChannelBitrate_63[0] = (Word16) tmp;
    for (i=1;i<2;i++) {
      fgets( InputLine, 100, fp);
      sscanf(InputLine, "%d", &tmp);
      ChannelBitrate_63[i] = (Word16) tmp;
    }

    flag=0;
    while(flag == 0 ) {
      fgets( InputLine, 100, fp);
      if( InputLine[0] != '#') flag=1;
    }
    sscanf(InputLine, "%d", &tmp);
    ChannelBitrate_SID[0] = (Word16) tmp;
    for (i=1;i<2;i++) {
      fgets( InputLine, 100, fp);
      sscanf(InputLine, "%d", &tmp);
      ChannelBitrate_SID[i] = (Word16) tmp;
    }

    fclose(fp);
    return;
}


static int Strincmp( const char *s, const char *t, size_t max )
{
    for( ; max > 1; ++s, ++t, --max ) {
        if( toupper( *s ) != toupper( *t ))
            break;
        if( *s == '\0' )
            return( 0 );
    }
    return( toupper( *s ) - toupper( *t ));
}


static FILE *OpenBinfile( char *name, char *mode )
{
    FILE *fp;

    if( toupper( *mode ) == 'W' ) { /* Write access */
        if(( fp = fopen( name, OPEN_WB )) == NULL ) { 
            printf( "Can't open output file '%s'\n", name );
            exit( 1 );
        }
    } else { /* Read access */
        if(( fp = fopen( name, OPEN_RB )) == NULL ) { 
            printf( "Can't open file '%s'\n", name );
            exit( 1 );
        } 
    }
    return( fp );
}


/*___________________________________________________________________________
 |                                                                           |
 |     Main - Program                                                        |
 |                                                                           |
 |___________________________________________________________________________|
*/
int main( int argc, char *argv[] )
{
#ifdef VAX
    Word32  VaxWriteCnt=0;
#endif
    FILE    *bit_infile, *chan_outfile, *chanfile;
    Word32  framecount, frameno = 0;
    Word16  ChannelBitrate,NumChannelBytes,ChanMode,end_bsf,end_cbf;
    char    InfoBitStream[24], ChannelBitStream[200];
    char    outtext[] = " Encoding Frame: %4d\r";

    if( (argc < 5) || (argc > 6) ) {
        fprintf( stderr, "------------------------------------------------------------------------------\n");
        fprintf( stderr, "|Usage: CHAN_ENC InfoBitFile ChanBitFile ConfigFile ChanBitRate[-File] [/f=f]| \n");
        fprintf( stderr, "------------------------------------------------------------------------------\n");
        fprintf( stderr, "\n");
        fprintf( stderr, "   InfoBitFile - encoded G.723.1 bitstream file\n");
        fprintf( stderr, "   ChanBitFile - channel bitstream file\n");
        fprintf( stderr, "   ConfigFile  - configuration file with allowed channel bitrates\n");
        fprintf( stderr, "   ChanBitRate - either: desired channel bit-rate in bps (e.g. 11000)\n");
        fprintf( stderr, "                     or: file name with channel bit-rates\n");
        fprintf( stderr, "   /f=         - number of frames to be encoded (optional) \n");
        return( 1 );
    }

    bit_infile  = OpenBinfile( argv[1], "r");
    chan_outfile = OpenBinfile( argv[2], "w");


    if (isdigit(argv[4][0])){
       /* A fixed channel bitrate has been specified on the command line */
       /* -------------------------------------------------------------- */
       ChanMode = 0;
       ChannelBitrate = (Word16) atoi(argv[4]);
    }else{
       /* A file with individual channel bitrates for each frame */ 
       /* has been specified on the command line                 */
       /* ------------------------------------------------------ */
       ChanMode = 1;
       chanfile = OpenBinfile( argv[4], "r");
    }


    frameno = 0;
    if(argc==6 && !Strincmp( argv[5], "/f=", 3 )){
        frameno = atoi( &argv[5][3] );
        if( frameno < 0 ) {
             printf( "frame > 0\n" );
             return( 1 );
        }
    }
    if(argc==6 && !Strincmp( argv[5], "-TestSequence", 13 )){
       TESTSEQUENCE = True;
    }


    printf( "  _____________________________________________\n" );
    printf( " |                                             |\n" );
    printf( " |       G.723.1 Channel Encoder Simulation    |\n" );
    printf( " |               Version 3.1                   |\n" );
    printf( " |_____________________________________________|\n\n" );

    printf( "    Info Bitstream Input File          : %s\n", argv[1] );
    printf( "    Transmitted Channel Bitstream File : %s\n", argv[2] );
    printf( "    Channel Bitrate Configuration File : %s\n", argv[3] );
    if (ChanMode == 0)
       printf( "    Channel Bitrate                    : %d\n", ChannelBitrate );
    if (ChanMode == 1)
       printf( "    Channel Bitrate File               : %s\n", argv[4] );
    if( frameno == 0 )
       printf( "    Simulated Frames                   : whole file\n" );
    else
       printf( "    Simulated Frames                   : %d\n", frameno );
    if (TESTSEQUENCE == True) {
       printf( "\n" );
       printf( "    Caution:  You are in testing mode !!\n");
       printf( "    ------------------------------------\n");
    }
    printf( "\n" );


    /* Read file with allowed channel bitrates */
    /* --------------------------------------- */
    GetChannelBitrates(argv[3]);


    for( framecount = 0; framecount++ < frameno || frameno == 0; ) {

       /* Read one frame from bitstream file */
       /* ---------------------------------- */
       end_bsf = GetSpeechFrame (bit_infile, InfoBitStream);
       if (end_bsf == 1) break;


       if (((Word16) InfoBitStream[0] & 0x3) < 3) {

          /* ------------------------------- */
          /* The frame has to be transmitted */
          /* ------------------------------- */

          /* Read channel bitrate for actual frame */
          /* ------------------------------------- */
          if (ChanMode == 1) {
             end_cbf = fread((char*)&ChannelBitrate,sizeof(Word16),1,chanfile);
             if (end_cbf != 1) break;
          }

          printf( outtext, framecount+1 );

          /* Channel encoder */
          /* --------------- */
          ChannelEncoder( InfoBitStream, ChannelBitStream, ChannelBitrate, &NumChannelBytes);

       }else{

          /* ----------------------------------------------------------------------- */
          /* The frame is not transmitted. It is signalled to generate comfort noise */
          /* ----------------------------------------------------------------------- */

	  NumChannelBytes     = 2;
	  ChannelBitStream[0] = 0xAF; /* this is a unused BCH word */
	  ChannelBitStream[1] = 0x16; /* this is a unused BCH word */
      }  

#ifdef VAX
       VaxWriteCnt += NumChannelBytes;
#endif
       /* Write packed channel bitstream */
       /* ------------------------------ */
       fwrite ((char*)ChannelBitStream, sizeof(char),(size_t)NumChannelBytes,chan_outfile);

    }
    printf("\n done \n");
    fclose( bit_infile );

#ifdef VAX
    if ((VaxWriteCnt % 2) == 1) {
       /* Adapt to 16 bit boundary */
       /* ------------------------ */
       ChannelBitStream[0]=0;
       fwrite ((char*)ChannelBitStream, sizeof(char),1,chan_outfile);
    }
#endif
    fclose( chan_outfile );
    if (ChanMode == 1) fclose( chanfile );

    return( 0 );
}
